     ////////////////////////////////////////////////////////
    // THE COMPARATIVE CODE-UNDERSTANDING SYSTEM           //
   // Written by Andrew Broad                             ///
  //                                                     ////
 // AggregationMatchingPolicies.java (2001:12:10 14:54) /////
/////////////////////////////////////////////////////////////
// Methods of AggregationMatchingPolicies:             /////
// - matchVectors                                      ////
// - matchUnorderedVectors                             ///
// - matchOrderedVectors                               //
////////////////////////////////////////////////////////

import java.io.*;
import java.util.*;
import jex.*;

/*****************************************************************************
** The class `AggregationMatchingPolicies' implements various policies used
** for matching a pair of aggregations.
*****************************************************************************/

public class AggregationMatchingPolicies {
	/********************************************************************/
	// static AggregationMatchingPolicies.matchVectors :
	//        (Vector x Vector x byte x Object x Object) -> void
	/*********************************************************************
	** Matches two vectors, selecting an appropriate aggregation matching 
	** policy.
	** @param vector1 The first vector.
	** @param vector2 The second vector.
	** @param mode Which mode it's comparing in (CORRESPONDENCE or
	**             PLAUSIBLE_CORRESPONDENCE).
	** @param type The FIL type of the vectors (optional - can be null).
	** @param subsForSelf A value to substitute for SELF (optional - can
	**                    be null).
	** @param entityPair The current entity-pair for the summarised output
	**                   (optional).
	** @return True if the vectors match, false otherwise.
	*********************************************************************/

	public static boolean matchVectors (Vector vector1, Vector vector2,
	                                    byte mode, Object type,
	                                    Object subsForSelf,
	                                    EntityPair entityPair)
	{
		if (vector1 instanceof OrderedVector
		 && vector2 instanceof OrderedVector)
		{
			return matchOrderedVectors((OrderedVector)vector1,
			                           (OrderedVector)vector2,
			                       mode, subsForSelf, entityPair);
		}
		else {
			return matchUnorderedVectors(vector1, vector2, mode,
			                       type, subsForSelf, entityPair);
		}
	}

	/********************************************************************/
	// static AggregationMatchingPolicies.matchUnorderedVectors :
	//        (Vector x Vector x byte x Object x Object) -> boolean
	/*********************************************************************
	** Matches two unordered vectors. Matches every element in the first
	** vector to an element in the second, and vice versa.
	** @param vector1 The first unordered vector.
	** @param vector2 The second unordered vector.
	** @param mode Which mode it's comparing in (CORRESPONDENCE or
	**             PLAUSIBLE_CORRESPONDENCE).
	** @param type The FIL type of the vectors (optional - can be null).
	**             This is used to determine whether the match should be
	**             cut once the first pair of matching frames is found
	**             (if it's an aggregation of frames with slot "name").
	** @param subsForSelf A value to substitute for SELF (optional - can
	**                    be null).
	** @param entityPair The current entity-pair for the summarised output
	**                   (optional).
	** @return Whether the two vectors match.
	*********************************************************************/

	private static boolean matchUnorderedVectors (Vector vector1,
	                                              Vector vector2,
	                                              byte mode,
	                                              Object type,
	                                              Object subsForSelf,
	                                              EntityPair entityPair)
	{
		///////////////////////////////////////////////////////
		// If the aggregations are known to be fillers       //
		// of a slot of type aggregation of frame,           //
		// then if the frame class has a slot called "name", //
		// then the aggregation match is to be cut           //
		// when it finds the first pair of matching fillers. //
		///////////////////////////////////////////////////////

		boolean cutAfterFirstElementMatch= false;

		if (type instanceof Frame
		 && FrameClass.lookup("unordered_aggregation_type").
		      subsumes(((Frame)type).getFrameClass()))
		{
			FrameClass elementFrameClass= null;
			try {
				elementFrameClass= FrameClass.lookupX(
				 ((Frame)((Frame)type).gsf("element_type")).
				 gsf("name").toString());
				Vector elementFrameClassSlots
				= elementFrameClass.getSlots();
				ListIterator eFCslots= elementFrameClassSlots.
				                       listIterator();
				while (eFCslots.hasNext() &&
				       ! cutAfterFirstElementMatch)
				{
					if (((Slot)eFCslots.next()).getName().
					    equals("name"))
					{
						cutAfterFirstElementMatch
						= true;
					}
				}
			} catch (FrameClassDoesNotExist fcdne) {}
		}

		boolean difference= false;

		//////////////////////////////////////////////////////
		// For each element in vector1,                     //
		// look for (a) corresponding element(s) in vector2 //
		//////////////////////////////////////////////////////

		ListIterator li1= vector1.listIterator();
		int i2= 0; // index into vector2
		if (vector2.size() == 0) {
			return (vector1.size() == 0);
		}

		Vector vector2clone= (Vector)vector2.clone();

		while (li1.hasNext()) {
			Object element1= li1.next();
			boolean matched= false; // found a matching element?
			int i2start= i2; // the start element in vector2
			do {
				Object element2= vector2.elementAt(i2);

				if (CCUS_FMA.matchFillers(element1, element2,
				         null, mode, subsForSelf, entityPair))
				{
					matched= true;
					vector2clone.remove(element2);
				}
				i2= (i2 + 1) % vector2.size();
				// next element in vector2
			} while (i2 != i2start &&
			         ! (matched && cutAfterFirstElementMatch));

			if (! matched) {
				CCUS_FMA.indent();
				System.out.print(element1);
				if (element1 instanceof Frame &&
				!FrameClass.lookup("higher_level_constraint").
				subsumes(((Frame)element1).getFrameClass())) {
					System.out.print(" ("
					+ ComparisonEngine._prettyPrint.
					generateDescription((Frame)element1,
					false) + ")");
					if (entityPair != null
					&& mode == CorrespondenceHeuristics.
					           CORRESPONDENCE)
					{
						entityPair.
						addNonCorrespondence
						((Frame)element1, 1);
					}
				}
				System.out.println(" doesn't correspond to an"
				+ "ything in the second model.");
				difference= true;
			}
		}

		/////////////////////////////////////////////////////////////
		// For each element in vector2 that was not matched in the //
		// above loop, look for a corresponding element in vector1 //
		/////////////////////////////////////////////////////////////

		ListIterator li2= vector2clone.listIterator();
		int i1= 0; // index into vector1
		if (vector1.size() == 0) {
			return (vector2clone.size() == 0);
		}

		while (li2.hasNext()) {
			Object element2= li2.next();
			boolean matched= false; // found a matching element?
			int i1start= i1; // the start element in vector1
			do {
				Object element1= vector1.elementAt(i1);

				if (CCUS_FMA.matchFillers(element1, element2,
				         null, mode, subsForSelf, entityPair))
				{
					matched= true;
				}
				i1= (i1 + 1) % vector1.size();
				// next element in vector1
			} while (i1 != i1start &&
			         ! (matched && cutAfterFirstElementMatch));

			if (! matched) {
				CCUS_FMA.indent();
				System.out.print(element2);
				if (element2 instanceof Frame &&
				!FrameClass.lookup("higher_level_constraint").
				subsumes(((Frame)element2).getFrameClass())) {
					System.out.print(" ("
					+ ComparisonEngine._prettyPrint.
					generateDescription((Frame)element2,
					false) + ")");
					if (entityPair != null
					&& mode == CorrespondenceHeuristics.
					           CORRESPONDENCE)
					{
						entityPair.
						addNonCorrespondence
					  	((Frame)element2, 2);
					}
				}
				System.out.println(" doesn't correspond to an"
				+ "ything in the first model.");
				difference= true;
			}
		}
		return ! difference;
	}

	/********************************************************************/
	// static AggregationMatchingPolicies.matchOrderedVectors :
	//        (OrderedVector x OrderedVector x byte x Object) -> boolean
	/*********************************************************************
	** Matches two ordered vectors. Requires the vectors to be the same
	** size, and traverses them in parallel, matching elements in
	** corresponding positions. This is the first of three ordered-vector
	** aggregation matching policies identified on Notebook VI p.847.
	** @param vector1 The first ordered vector.
	** @param vector2 The second ordered vector.
	** @param mode Which mode it's comparing in (CORRESPONDENCE or
	**             PLAUSIBLE_CORRESPONDENCE).
	** @param subsForSelf A value to substitute for SELF (optional - can
	**                    be null).
	** @param entityPair The current entity-pair for the summarised output
	**                   (optional).
	** @return Whether the two vectors match.
	*********************************************************************/

	private static boolean matchOrderedVectors (OrderedVector vector1,
	                                            OrderedVector vector2,
	                                            byte mode,
	                                            Object subsForSelf,
	                                            EntityPair entityPair)
	{
		boolean difference= false;
		if (vector1.size() != vector2.size()) {
			difference= true;
		}

		ListIterator li1= vector1.listIterator();
		ListIterator li2= vector2.listIterator();
		while (li1.hasNext() && li2.hasNext()) {
			Object element1= li1.next();
			Object element2= li2.next();

			if (! CCUS_FMA.matchFillers(element1, element2, null,
			                       mode, subsForSelf, entityPair))
			{
				CCUS_FMA.indent();
				System.out.print(element1);
				if (element1 instanceof Frame &&
				!FrameClass.lookup("higher_level_constraint").
				subsumes(((Frame)element1).getFrameClass())) {
					System.out.print(" ("
					+ ComparisonEngine._prettyPrint.
					generateDescription((Frame)element1,
					false) + ")");
				}
				System.out.print(" doesn't correspond to "
				+ element2);
				if (element2 instanceof Frame &&
				!FrameClass.lookup("higher_level_constraint").
				subsumes(((Frame)element2).getFrameClass())) {
					System.out.print(" ("
					+ ComparisonEngine._prettyPrint.
					  generateDescription
					  ((Frame)element2, false) + ")");
				}
				System.out.println(".");
				difference= true;
			}
		}
		return ! difference;
	}
}
