      ////////////////////////////////////////
     // THE CONSTRAINT-UNDERSTANDING SYSTEM //
    //  (Version 2)                        ///
   // Written by Andrew Broad             ////
  //                                     /////
 // Case.java (2000:12:11 16:10)        //////
//////////////////////////////////////////////
// Methods of Case:                    ///////
// - Case constructor                  ///////
// - getCaseFrame                      ///////
// - getProblemFrameClass              ///////
// - match                             ///////
// - matchFrames                       //////
// - matchFillers                      /////
// - extract                           ////
// - extractFrame                      ///
// - extractFiller                     //
////////////////////////////////////////

import java.io.*;
import java.util.*;
import jex.*;

/*****************************************************************************
** The class `Case' represents a case, which encapsulates a "case" frame. Case
** is the unit of matching and adaptation.
*****************************************************************************/

public class Case implements Serializable {
	private Frame _caseFrame;
	private HashtableStack _frameMap;
	private HashtableStack _variableMap;
	private Vector _result;

	/********************************************************************/
	// Case constructor : Frame ->
	/*********************************************************************
	** Constructs a case from a given frame. Marks all frames in the
	** transitive closure of the problem slot as being such.
	** @param caseFrame A frame to construct the case from, which should
	**  be of the "case" frame class. All frames in the transitive closure
	**  of this one's slots are treated as part of the case.
	*********************************************************************/

	public Case (Frame caseFrame) {
		_caseFrame= caseFrame;
		try {
			((Frame)caseFrame.getSlotFiller("problem")).
				markCaseProblemPart();
		} catch (InvalidSlot x) {
			System.err.println("InvalidSlot caught in Case " +
			"constructor!\nThis means that the case frame doesn't"
			+ " have a `problem' slot");
			caseFrame.print(System.err);
			System.exit(1);
		}
	}

	/********************************************************************/
	// Case.getCaseFrame : void -> Frame
	/*********************************************************************
	** Delivers up the underlying frame of this case.
	** @return The case representing this case.
	*********************************************************************/
 
	public Frame getCaseFrame () {
		return _caseFrame;
	}

	/********************************************************************/
	// Case.getProblemFrameClass : void -> FrameClass
	/*********************************************************************
	** Delivers up the frame class of the frame which fills this case's
	** `problem' slot.
	** @return The frame class of this case's top-level `problem' frame.
	*********************************************************************/

	public FrameClass getProblemFrameClass () {
		try {
			return ((Frame)getCaseFrame().
			        getSlotFiller("problem")).getFrameClass();
		} catch (InvalidSlot is) {
			System.err.println("Case does not have a \"problem\" "
			+ "slot!");
			getCaseFrame().print(System.err);
			System.exit(1); return null;
		}
	}

	/********************************************************************/
	// Case.match : Frame -> boolean
	/*********************************************************************
	** Matches this case with the given frame, returning true if the match
	** is good enough to extract the HLC suggested by the case.
	** @param frame A frame to match the `problem' part of this case to.
	**  All slots in the case's `problem' frame will be matched; any 
	**  frames will be matched recursively.
	** @return Whether `frame' matches the case well enough.
	*********************************************************************/

	public boolean match (Frame frame) {
		_frameMap= new HashtableStack();
		_variableMap= new HashtableStack();
		try {
			if (! matchFrames
			   ((Frame)_caseFrame.getSlotFiller("problem"),frame))
			{
				_frameMap= null; _variableMap= null;
				return false;
			} else return true;
		} catch (InvalidSlot x) {
			System.err.println
			           ("Invalid case - no `problem' slot!");
			System.exit(1); return false;
		}
	}

	/********************************************************************/
	// Case.matchFrames : Frame x Frame -> boolean
	/*********************************************************************
	** Matches a frame in the case with a frame in the problem situation.
	** @param caseFrame The frame to match in the case.
	** @param sitFrame The frame to match in the problem situation.
	** @return Whether the problem situation matches (all the slots in
	**         the) case frame.
	*********************************************************************/

	private boolean matchFrames (Frame caseFrame, Frame sitFrame) {

		//////////////////////////////////////////////////////////////
		// Match fails if case frame already corresponds to a diff- //
		// erent frame in the situation; if the frames have been    //
		// matched already, then cut to avoid infinite recursion.   //
		//////////////////////////////////////////////////////////////
		Frame assFrame;
		if ((assFrame= (Frame)_frameMap.lookup(caseFrame)) != null)
		 	return (assFrame == sitFrame);

		_frameMap.store(caseFrame,sitFrame);

		/////////////////////////////////////////////////////////////
		// Compare frame classes: match fails if situation frame's //
		// is not a subclass of the case frame's                   //
		/////////////////////////////////////////////////////////////

		if (! caseFrame.getFrameClass().subsumes
		     (sitFrame.getFrameClass()))
			return false;
			
		///////////////////////////////////////////////////////////
		// Match each slot filled in the case frame with the     //
		// corresponding slot in the situation frame (only those //
		// slots which were set directly in the case frame)      //
		///////////////////////////////////////////////////////////

		ListIterator slotsInCaseFrame= caseFrame.getSlots();
		while (slotsInCaseFrame.hasNext()) {
			Slot slot= (Slot)slotsInCaseFrame.next();

			/////////////////////////////////
			// Whole match fails           //
			// if one slot fails to match! //
			/////////////////////////////////
			if (caseFrame.isSlotSetDirectly(slot)
			 && ! matchFillers (caseFrame.getSlotFiller(slot),
			                     sitFrame.getSlotFiller(slot)))
			{
				if (ExtractionEngine._showIntermediateOutput){
					System.out.println("match failed (on "
					+ caseFrame + "." + slot + ").");
				}
				return false;
			}
		}

		////////////////////////////////////////////////////////////
		// Match succeeds if it comes through all that unscathed! //
		////////////////////////////////////////////////////////////

		return true;
	}

	/********************************************************************/
	// Case.matchFillers : Object x Object -> boolean
	/*********************************************************************
	** Matches a slot filler in the case to one in the problem situation.
	** @param caseSlotFiller The slot filler from the case.
	** @param sitSlotFiller The slot filler from the problem situation.
	** @return Whether case value matches the problem situation value.
	*********************************************************************/

	private boolean matchFillers (Object caseSlotFiller,
	                              Object  sitSlotFiller)
	{
		//////////////////////////////////////////////////////////////
		// If the case slot filler is a variable, bind it to the    //
		// corresponding situation slot filler or, if it is bound   //
		// already, match its binding to the situation slot filler. //
		// As of Version 2, variables may be bound to null: such a  //
		// binding is represented as an instance of Null to         //
		// distinguish it from unbound variables.                   //
		//////////////////////////////////////////////////////////////

		if (caseSlotFiller instanceof Variable) {
			Object binding= _variableMap.
				lookup(caseSlotFiller.toString());
			if (binding == null) {
				_variableMap.store(caseSlotFiller.toString(),
				                   sitSlotFiller == null
				                   ? new Null()
				                   : sitSlotFiller);
				return true;
			} else {
				return matchFillers(binding,sitSlotFiller);
			}
		}

		///////////////////////////////////////////////////////////
		// If value in case is an instance of Null then value in //
		// situation must be null; in all other circumstances,   //
		// value in situation must be non-null.                  //
		///////////////////////////////////////////////////////////

		if (caseSlotFiller instanceof Null) {
			return (sitSlotFiller == null);
		}
		if (sitSlotFiller == null) {
			return false;
		}

		//////////////////////////////////////////////////////////////
		// Match fails if sitSlotFiller is not Java-type-compatible //
		// with caseSlotFiller (unless caseSlotFiller's a variable) //
		//////////////////////////////////////////////////////////////

		if (! caseSlotFiller.getClass().isInstance(sitSlotFiller)) {
			return false;
		}

		////////////////////////////////////////////////
		// If they're frames, match them recursively. //
		////////////////////////////////////////////////

		if (caseSlotFiller instanceof Frame) {
			return matchFrames((Frame)caseSlotFiller,
			                   (Frame)sitSlotFiller);
		}

		//////////////////////////////////////////
		// If they're defined type instances,   //
		// compare type then underlying values. //
		//////////////////////////////////////////

		if (caseSlotFiller instanceof DefinedTypeInstance) {
			return (((DefinedTypeInstance)caseSlotFiller).
			         getDefinedType()
			     == ((DefinedTypeInstance) sitSlotFiller).
			         getDefinedType())
			&&
			matchFillers
			(((DefinedTypeInstance)caseSlotFiller).getValue(),
			 ((DefinedTypeInstance)sitSlotFiller).getValue());
		}

		//////////////////////////////////////////////////////////////
		// If they're unordered vectors, match every element in the //
		// case set to an element in the prompt set.                //
		//////////////////////////////////////////////////////////////

		if (caseSlotFiller instanceof UnorderedVector) {
			Enumeration caseSet= ((UnorderedVector)
			                      caseSlotFiller).elements();
			while (caseSet.hasMoreElements()) {
				Object caseElement= caseSet.nextElement();
				boolean matched= false;
				Enumeration sitSet= ((UnorderedVector)
				                    sitSlotFiller).elements();
				while (sitSet.hasMoreElements()) {
					Object sitElement=
						sitSet.nextElement();
					_frameMap.push();
					_variableMap.push();
					if (matchFillers
					    (caseElement,sitElement))
					{
						matched= true; break;
					} else {
						_frameMap.pop();
						_variableMap.pop();
					}
				}
				if (! matched) return false;
			}
			return true;
		}

		////////////////////////////////////////////////////////////
		// If they're ordered vectors, match every element in     //
		// case list to the corresponding element in prompt list. //
		////////////////////////////////////////////////////////////

		if (caseSlotFiller instanceof OrderedVector) {
			if (((OrderedVector)caseSlotFiller).size() !=
		            ((OrderedVector)sitSlotFiller).size())
				return false;
			Enumeration caseList= ((OrderedVector)caseSlotFiller).
			                       elements();
			Enumeration sitList= ((OrderedVector)sitSlotFiller).
			                      elements();
			while (caseList.hasMoreElements()) {
				if (! matchFillers(caseList.nextElement(),
				                    sitList.nextElement()))
					return false;
			}
			return true;
		}
		
		/////////////////////////////////////
		// By default, call Object.equals. //
		/////////////////////////////////////

		return caseSlotFiller.equals(sitSlotFiller);
	}

	/********************************************************************/
	// Case.extract : void -> Vector<Frame>
	/*********************************************************************
	** Extract this case's solution frames into new instantiations for the
	** problem situation (i.e. the input model).
	** The client must have called match first, to check that it is
	** appropriate to extract the solution and to obtain the bindings for
	** variables and frames in the case. This method can be used when the
	** case is for proposing a ballpark solution, because it automatically
	** creates a new solution frame.
	** If it is an adaptation case, use extractFrame to adapt an existing
	** solution frame.
	** @return The (usually singleton) vector of solution frames thus
	**         created.
	*********************************************************************/

	public Vector extract () {
		_result= new Vector();
		Object solution= null;
		try {
			solution= _caseFrame.getSlotFiller("solution");
		} catch (InvalidSlot x) {
			System.err.println("InvalidSlot caught in "
			+ "Case.extract!");
			System.exit(1);
		}
		if (solution instanceof Frame) {
			_result.addElement(extractFrame((Frame)solution));
		} else if (solution instanceof Vector) {
			Enumeration solutionFrames
			= ((Vector)solution).elements();
			while (solutionFrames.hasMoreElements()) {
				_result.addElement(extractFrame((Frame)
				       solutionFrames.nextElement()));
			}
		} else {
			System.err.println("Case.extract: solution slot "
			+ "filler is not either a frame or a set of frames!");
			System.exit(1);
		}
		_frameMap= null; _variableMap= null;
		return _result;
	}

	/********************************************************************/
	// Case.extractFrame : Frame -> Frame
	/*********************************************************************
	** Create a new frame, copying from the given frame from the case.
	** The client must have called match first, to check that it is
	** appropriate to extract the solution and to obtain the bindings for
	** variables and frames in the case. This method can be used whether
	** the case is for proposing a ballpark solution or for adaptation.
	** @param caseSolutionFrame The solution frame from the case.
	** @return A solution frame for the problem situation.
	*********************************************************************/

	public Frame extractFrame (Frame caseSolutionFrame) {
		if (_frameMap == null) {
			System.err.println("Case.extractFrame cannot be "
			+ "called now - this case has not been matched "
			+ "successfully, or has not been matched at all.");
			System.exit(1);
		}
		
		Frame newSolutionFrame= new Frame(caseSolutionFrame.
			                          getFrameClass());

		//////////////////////////////////////////////////////////////
		// Put the new solution frame in _frameMap so it will be    //
		// matched rather than recreated if it is encountered again //
		//////////////////////////////////////////////////////////////

		_frameMap.store(caseSolutionFrame,newSolutionFrame);

		//////////////////////////////////////////////////////////////
		// Copy each slot in the solution frame into the target     //
		// frame, substituting any frames that have been matched to //
		// frames in the problem situation, and substituting any    //
		// variables with the values to which they are bound.       //
		//////////////////////////////////////////////////////////////
		ListIterator solutionSlots= caseSolutionFrame.getSlots();
		while (solutionSlots.hasNext()) {
			Slot slot= (Slot)solutionSlots.next();
			Object caseFiller= caseSolutionFrame.
			                   getSlotFiller(slot);
			Object filler= extractFiller(caseFiller);
			if (filler instanceof OrderedVector
			 && slot.isUnorderedAggType())
			{
				UnorderedVector newFiller
				= new UnorderedVector();
				ListIterator li= ((OrderedVector)filler).
				                 listIterator();
				while (li.hasNext()) {
					newFiller.addElement(li.next());
				}
				filler= newFiller;
			}
			newSolutionFrame.setSlotFiller(slot,filler);
		}
		return newSolutionFrame;
	}

	/********************************************************************/
	// Case.extractFiller : Object -> Object
	/*********************************************************************
	** Process a given slot filler from a solution frame for extraction
	** when instantiating that solution frame for the model. Matched
	** frames are replaced with the frames they matched in the situation,
	** and bound variables are replaced with their bindings.
	** @param filler A slot filler to process.
	** @return The result of processing the slot filler as described
	**         above.
	*********************************************************************/

	private Object extractFiller (Object filler) {
		Object subst;

		/////////////////////////////////////////////////////////////
		// Replace matched frames from the case with the frames    //
		// they have matched in the situation. If a frame is not   //
		// associated with a frame in the situation, then create a //
		// new frame and copy the case frame into it recursively.  //
		/////////////////////////////////////////////////////////////

		if (filler instanceof Frame) {
			if ((subst= _frameMap.lookup(filler)) != null)
				return subst;
			else {
				Frame f= extractFrame((Frame)filler);
				_result.addElement(f);
				return f;
			}
		}

		//////////////////////////////////////////////////////////////
		// Replace bound variables with their bindings.             //
		// A variable that is bound to null has an instance of Null //
		// as its binding, to distinguish it from unbound variables //
		//////////////////////////////////////////////////////////////

		else if (filler instanceof Variable) {
			Object binding=_variableMap.lookup(filler.toString());
			if (binding instanceof Null)
				return null;
			if (binding != null)
				return binding;
			else {
				System.err.println("Unbound variable "
				+ filler + "!");
				return null; // just to please javac! :-)
			}
		}

		//////////////////////////////////////////////////////////////
		// If it's an aggregation, process the elements recursively //
		//////////////////////////////////////////////////////////////

		else if (filler instanceof Vector) {
			Vector result;
			if (filler instanceof UnorderedVector)
				result= new UnorderedVector();
			else
				result= new OrderedVector();

			Enumeration elements= ((Vector)filler).elements();
			while (elements.hasMoreElements()) {
				result.addElement
				      (extractFiller(elements.nextElement()));
			}
			return result;
		}

		//////////////////////////////////////////////////////
		// If filler is anything else, plug it in as it is. //
		//////////////////////////////////////////////////////

		else return filler;
	}
}
