      ////////////////////////////////////////
     // THE CONSTRAINT-UNDERSTANDING SYSTEM //
    //  (Version 2)                        ///
   // Written by Andrew Broad             ////
  //                                     /////
 // EntityPair.java (2002:01:07 15:02)  //////
//////////////////////////////////////////////
// Methods of EntityPair:              ///////
// - EntityPair                        ///////
// - addCorrespondence                 //////
// - addNonCorrespondence              /////
// - addDifference                     ////
// - printHeaderLine                   ///
// - generateOutput                    //
////////////////////////////////////////

import java.io.*;
import java.util.*;

/*****************************************************************************
** The class `EntityPair' manages the summarised output for a pair of
** entities.
*****************************************************************************/

public class EntityPair {
	private Frame _entity1;
	private Frame _entity2;
	private Vector/*Correspondence*/ _correspondences= new Vector();
	private Vector/*Correspondence*/ _HLCcorrespondences= new Vector();
	private Vector/*NonCorrespondence*/ _nonCorrespondences1=new Vector();
	private Vector/*NonCorrespondence*/ _nonCorrespondences2=new Vector();
	private Vector/*Difference*/ _differences= new Vector();

	private static int _numberOfInstances= 0;
	private int _instanceNumber;

	/********************************************************************/
	// EntityPair constructor : void ->
	/*********************************************************************
	** Constructs an empty entity-pair.
	** @param name1 The name of the first entity.
	** @param name2 The name of the second entity.
	*********************************************************************/
 
 	public EntityPair (Frame entity1, Frame entity2) {
 		_entity1= entity1;
 		_entity2= entity2;
 		_instanceNumber= ++_numberOfInstances;
 		Summary.addEntityPair(this);
	}

	/********************************************************************/
	// EntityPair.addCorrespondence : Frame x Frame -> void
	/*********************************************************************
	** Adds a correspondence for this entity-pair.
	** @param frame1 The frame from the first model.
	** @param frame2 The frame from the second model.
	*********************************************************************/

	public void addCorrespondence (Frame frame1, Frame frame2) {
		boolean hlc1= FrameClass.lookup("higher_level_constraint").
		              subsumes(frame1.getFrameClass());
		boolean hlc2= FrameClass.lookup("higher_level_constraint").
		              subsumes(frame2.getFrameClass());
		if (hlc1 || hlc2) {
			_HLCcorrespondences.add
			                 (new Correspondence(frame1, frame2));
		} else {
			_correspondences.add
			                 (new Correspondence(frame1, frame2));
		}
	}

	class Correspondence {
		Frame _frame1;
		Frame _frame2;

		public Correspondence (Frame frame1, Frame frame2) {
			_frame1= frame1;
			_frame2= frame2;
		}

		public void generateOutput (PrintWriter summaryFile) {
			summaryFile.print("<LI><A HREF=\""
			+ _frame1.getFrameClass() + ".html#"
			+ _frame1.getNumber() + "\">");
			String f1d= null;
			if (FrameClass.lookup("higher_level_constraint").
		                       subsumes(_frame1.getFrameClass()))
		        {
				f1d= _frame1.getFrameClass() + "#"
				     + _frame1.getNumber();
		        } else {
				f1d= ComparisonEngine._prettyPrint.
		   	             generateDescription(_frame1, false);
		   	}
			summaryFile.print("<TT>" + f1d + "</TT></A>");

			summaryFile.print(" corresponds to ");

			Frame cf=
			Correspondences.correspondenceFrame(_frame1, _frame2);
			if (cf != null && cf.gsf("semantic_equivalence")
			                  instanceof DefinedTypeInstance)
			{
				DefinedTypeInstance se= (DefinedTypeInstance)
				               cf.gsf("semantic_equivalence");
				if (se.toString().equals("frame1_stronger")) {
					summaryFile.print("(is semantically "
					+ "stronger than) ");
				}
				if (se.toString().equals("frame2_stronger")) {
					summaryFile.print("(is semantically "
					+ "weaker than) ");
				}
			}

			summaryFile.print("<A HREF=\""
			+ _frame2.getFrameClass() + ".html#"
			+ _frame2.getNumber() + "\">");
			String f2d= null;
			if (FrameClass.lookup("higher_level_constraint").
		                       subsumes(_frame2.getFrameClass()))
		        {
				f2d= _frame2.getFrameClass() + "#"
				     + _frame2.getNumber();
		        } else {
				f2d= ComparisonEngine._prettyPrint.
		   	             generateDescription(_frame2, false);
		   	}
			summaryFile.print("<TT>" + f2d + "</TT></A>");
			summaryFile.println("</LI>");
		}
	}

	/********************************************************************/
	// EntityPair.addNonCorrespondence : Frame x byte -> void
	/*********************************************************************
	** Adds a non-correspondence for this entity-pair.
	** @param frame The frame from one model.
	** @param model Which model the frame is from (1 or 2).
	*********************************************************************/

	public void addNonCorrespondence (Frame frame, int model) {
		if (model == 1) {
			_nonCorrespondences1.add(
			  new NonCorrespondence(frame, 1));
		} else {
			_nonCorrespondences2.add(
			  new NonCorrespondence(frame, 2));		
		}
	}

	class NonCorrespondence {
		Frame _frame;
		int _model;

		public NonCorrespondence (Frame frame, int model) {
			_frame= frame;
			_model= model;
		}

		public void generateOutput (PrintWriter summaryFile) {
			summaryFile.print("<LI><A HREF=\""
			+ _frame.getFrameClass() + ".html#"
			+ _frame.getNumber() + "\">");
			String fd= ComparisonEngine._prettyPrint.
		   	            generateDescription(_frame, false);
			summaryFile.print("<TT>" + fd + "</TT></A>");
			summaryFile.print(" doesn't correspond to anything in"
			+ " the " + (_model == 1 ? "second" : "first")
			+ " model</LI>");
		}
	}

	/********************************************************************/
	// EntityPair.addDifference : Slot x Frame x Object x Frame x Object
	//                            -> void
	/*********************************************************************
	** Adds a difference for this entity-pair.
	** @param slot The slot for which there is a difference.
	** @param frame1 The frame from the first model.
	** @param filler1 The filler of slot for frame1.
	** @param frame2 The frame from the second model.
	** @param filler2 The filler of slot for frame2.
	*********************************************************************/

	public void addDifference (Slot slot, Frame frame1, Object filler1,
	                                      Frame frame2, Object filler2)
	{
		_differences.add(new Difference(slot, frame1, filler1,
		                                      frame2, filler2));
	}

	class Difference {
		Slot _slot;
		Frame _frame1;
		String _filler1;
		Frame _frame2;
		String _filler2;

		public Difference (Slot slot, Frame frame1, Object filler1,
	                                      Frame frame2, Object filler2)
		{
			_slot= slot;
			_frame1= frame1;
			_filler1= filler1 == null
			          ? "null"
			          : (filler1 instanceof FILisable
			             ? ((FILisable)filler1).FILise(true)
			             : filler1.toString());
			_frame2= frame2;
			_filler2= filler2 == null
			          ? "null"
			          : (filler2 instanceof FILisable
			             ? ((FILisable)filler2).FILise(true)
			             : filler2.toString());
		}

		public void generateOutput (PrintWriter summaryFile) {
			summaryFile.print("<LI>Difference in slot <TT>"
			+ _slot.getFrameClassDeclaredIn() + "." + _slot
			+ "</TT>:");
			summaryFile.println("<UL><LI><TT>"
			+ _frame1.FILise(true) + "." + _slot
			+ "</TT> = <TT>" + _filler1 + "</TT></LI>");
			summaryFile.println("<LI><TT>"
			+ _frame2.FILise(true) + "." + _slot
			+ "</TT> = <TT>" + _filler2 + "</TT></LI></UL></LI>");
		}
	}

	/********************************************************************/
	// EntityPair.printHeaderLine : PrintWriter -> void
	/*********************************************************************
	** Prints the HTML header-line for this entity-pair.
	** @param summaryFile The PrintWriter object for summary.html.
	*********************************************************************/

	public void printHeaderLine (PrintWriter summaryFile)
	{
		////////////
		// Header //
		////////////
		summaryFile.println("<LI><A HREF=\"#" + _instanceNumber
		+ "\"><TT>" + _entity1.gsf("name") + "</TT> corresponds to "
		+ "<TT>" + _entity2.gsf("name") + "</TT></A></LI>");
	}

	/********************************************************************/
	// EntityPair.generateOutput : PrintWriter -> void
	/*********************************************************************
	** Generates the HTML output for this entity-pair.
	** @param summaryFile The PrintWriter object for summary.html.
	*********************************************************************/

	public void generateOutput (PrintWriter summaryFile)
	{
		////////////
		// Header //
		////////////
		summaryFile.println("<H2><A NAME=\"" + _instanceNumber
		+ "\"></A><A HREF=\"entity_type.html#" + _entity1.getNumber()
		+ "\"><TT>" + _entity1.gsf("name") + "</TT></A> corresponds "
		+ "to <A HREF=\"entity_type.html#" + _entity2.getNumber()
		+ "\"><TT>" + _entity2.gsf("name") + "</TT></A></H2>");

		/////////////////////
		// Correspondences //
		/////////////////////
		ListIterator correspondences= _correspondences.listIterator();
		if (correspondences.hasNext()) {
			summaryFile.println("<UL>");
			while (correspondences.hasNext()) {
				((Correspondence)correspondences.next()).
				                 generateOutput(summaryFile);
			}
			summaryFile.println("</UL>");
		}

		//////////////////////////////////////
		// Non-Correspondences: first model //
		//////////////////////////////////////
		ListIterator nc= _nonCorrespondences1.listIterator();
		if (nc.hasNext()) {
			summaryFile.println("<UL>");
			while (nc.hasNext()) {
				((NonCorrespondence)nc.next()).generateOutput
				                               (summaryFile);
			}
			summaryFile.println("</UL>");
		}

		///////////////////////////////////////
		// Non-Correspondences: second model //
		///////////////////////////////////////
		nc= _nonCorrespondences2.listIterator();
		if (nc.hasNext()) {
			summaryFile.println("<UL>");
			while (nc.hasNext()) {
				((NonCorrespondence)nc.next()).generateOutput
				                               (summaryFile);
			}
			summaryFile.println("</UL>");
		}

		/////////////////
		// Differences //
		/////////////////
		ListIterator differences= _differences.listIterator();
		if (differences.hasNext()) {
			summaryFile.println("<UL>");
			while (differences.hasNext()) {
				((Difference)differences.next()).
				             generateOutput(summaryFile);
			}
			summaryFile.println("</UL>");
		}

		/////////////////////////
		// HLC-Correspondences //
		/////////////////////////
		correspondences
		= _HLCcorrespondences.listIterator();
		if (correspondences.hasNext()) {
			summaryFile.println("HLC-correspondences:");
			summaryFile.println("<UL>");
			while (correspondences.hasNext()) {
				((Correspondence)correspondences.next()).
				                 generateOutput(summaryFile);
			}
			summaryFile.println("</UL>");
		}
	}
}
