     ////////////////////////////////////////////////
    // THE COMPARATIVE CODE-UNDERSTANDING SYSTEM   //
   // Written by Andrew Broad                     ///
  //                                             ////
 // HLC_CCUS.java (2001:12:06 20:43)            /////
/////////////////////////////////////////////////////
// Methods of HLC_CCUS:                        //////
// - extractHLCs                               //////
// - forget1                                   //////
// - forget2                                   /////
// - submitExtractionPrompts(Frame,Slot)       ////
// - submitExtractionPrompts(Frame,Frame,Slot) ///
// - mergeHashtables                           //
////////////////////////////////////////////////

import java.io.*;
import java.util.*;
import jex.*;

/*****************************************************************************
** The class `HLC_CCUS' implements the part of the Comparative Constraint-
** Understanding System that co-ordinates the extraction of HLCs from the two
** models, and the comparison of pairs of HLCs thus extracted.
*****************************************************************************/

public class HLC_CCUS {
	private static int _numberOfHLCsExtracted0;
	public static int _numberOfHLCsExtracted1;
	public static int _numberOfHLCsExtracted2;

	///////////////////////////////////////////////////////////
	// HLC-remembrance structures: first model, second model //
	///////////////////////////////////////////////////////////
	private static OrderedHashtable/*FrameClass->Vector*/ _r1
	= new OrderedHashtable();
	private static OrderedHashtable/*FrameClass->Vector*/ _r2
	= new OrderedHashtable();

	/********************************************************************/
	// HLC_CCUS.extractHLCs : (Frame x Frame x Slot) -> void
	/*********************************************************************
	** Called by CCUS_FMA.processDifference to extract HLCs from
	** corresponding parts of a model.
	** @param frame1 A frame from the first model to extract HLCs from.
	** @param frame1 A frame from the second model to extract HLCs from.
	** @param slot The slot for which a difference was found.
	** @param entityPair The current entity-pair for the summarised output
	**                   (optional).
	*********************************************************************/

	public static void extractHLCs (Frame frame1, Frame frame2, Slot slot,
	                                EntityPair entityPair)
	{
		///////////////////////////////////
		// Extract HLCs from both models //
		///////////////////////////////////
		OrderedHashtable h1= null, h2= null;
		if (CCUS._considerExtractionPromptsTogether) {
			OrderedHashtable[] pairOfHashtables
			= submitExtractionPrompts(frame1,frame2,slot);
			h1= pairOfHashtables[0];
			h2= pairOfHashtables[1];
		} else {
			h1= submitExtractionPrompts(frame1,slot);
			_numberOfHLCsExtracted1+= _numberOfHLCsExtracted0;

			h2= submitExtractionPrompts(frame2,slot);
			_numberOfHLCsExtracted2+= _numberOfHLCsExtracted0;
		}

		//////////////////////////////////
		// Remember HLCs for comparison //
		//////////////////////////////////
		OrderedHashtable r1= h1;
		OrderedHashtable r2= h2;
		OrderedHashtable old_r1= (OrderedHashtable)_r1.clone();

		if (CCUS._rememberHLCs) {
			r1= mergeHashtables(_r1, h1);
			r2= mergeHashtables(_r2, h2);
		}

		//////////////////
		// Compare HLCs //
		//////////////////
		ListIterator keys= r1.keyListIterator();
		while (keys.hasNext()) {
			Object key= keys.next();
			Vector old_r1v= (Vector)old_r1.get(key);
			Vector r1v= (Vector)r1.get(key);
			Vector r2v= (Vector)r2.get(key);
			Vector h1v= (Vector)h1.get(key);
			Vector h2v= (Vector)h2.get(key);

			/////////////////////////////////
			// Without remembrance of HLCs //
			/////////////////////////////////
			if (!CCUS._rememberHLCs && r1v != null && r2v != null)
			{
				AggregationMatchingPolicies.matchVectors
				  ((Vector)r1v.clone(), (Vector)r2v.clone(),
				   CorrespondenceHeuristics.
				   PLAUSIBLE_CORRESPONDENCE,
				   null, null, entityPair);
			}

			//////////////////////////////
			// With remembrance of HLCs //
			//////////////////////////////
			if (CCUS._rememberHLCs) {
				if (h1v != null && r2v != null) {
					AggregationMatchingPolicies.
					  matchVectors
					  (h1v, (Vector)r2v.clone(),
					   CorrespondenceHeuristics.
					   PLAUSIBLE_CORRESPONDENCE,
					   null, null, entityPair);
				}
				if (old_r1v != null && h2v != null) {
					AggregationMatchingPolicies.
					  matchVectors
					  ((Vector)old_r1v.clone(), h2v,
					   CorrespondenceHeuristics.
					   PLAUSIBLE_CORRESPONDENCE,
					   null, null, entityPair);
				}
			}
		}
	}

	/********************************************************************/
	// HLC_CCUS.forget1 : Frame -> void
	/*********************************************************************
	** Remove an HLC from the remembrance structure for the
	** <B>first</B> model.
	** @param hlc An HLC from the first model to forget.
	*********************************************************************/

	public static void forget1 (Frame hlc) {
		Vector v= (Vector)_r1.get(hlc.getFrameClass());
		if (v != null) {
			v.removeElement(hlc);
			if (v.size() == 0) {
				_r1.remove(hlc.getFrameClass());
			}
		}
	}

	/********************************************************************/
	// HLC_CCUS.forget2 : Frame -> void
	/*********************************************************************
	** Remove an HLC from the remembrance structure for the
	** <B>second</B> model.
	** @param hlc An HLC from the second model to forget.
	*********************************************************************/

	public static void forget2 (Frame hlc) {
		Vector v= (Vector)_r2.get(hlc.getFrameClass());
		if (v != null) {
			v.removeElement(hlc);
			if (v.size() == 0) {
				_r2.remove(hlc.getFrameClass());
			}
		}
	}

	/********************************************************************/
	// HLC_CCUS.submitExtractionPrompts : (Frame x Slot)
	//                                    -> OrderedHashtable
	/*********************************************************************
	** Submits a given frame as prompt to the CUS for it to extract HLCs -
	** I'm assuming for the moment that HLCs can be extracted from each
	** model independently of the other - otherwise two frames
	** (corresponding frames from each model) will have to be considered
	** together.
	** @param frame A frame to extract HLCs from.
	** @param slot The slot for which a difference was found.
	** @return A hashtable of the HLCs extracted (each key is a frame
	**         class, and each value is a vector of the HLCs of that
	**         frame class).
	*********************************************************************/

	private static OrderedHashtable submitExtractionPrompts (Frame frame,
	                                                         Slot slot)
	{
		_numberOfHLCsExtracted0= 0;
		Object slotFiller= frame.getSlotFiller(slot);

		/////////////////////////////////////////////////////////////
		// If the slot filler is null, don't submit any frames as  //
		// extraction prompts                                      //
		// (without this cut, null would be treated as a non-frame //
		// value and the containing frame would be submitted).     //
		/////////////////////////////////////////////////////////////

		if (slotFiller == null)
			return new OrderedHashtable();

		///////////////////////////////////////
		// If the slot filler is a frame,    //
		// submit it as an extraction prompt //
		///////////////////////////////////////
		if (slotFiller instanceof Frame) {
			CCUS_FMA.indent();
			System.out.println("Submitting " + slotFiller
			+ " as an extraction prompt");

			OrderedHashtable returnValue= ExtractionEngine.
			                       CCUSextract((Frame)slotFiller);
			_numberOfHLCsExtracted0= ExtractionEngine.
			                         _numberOfHLCsExtracted;
			return returnValue;
			
		} else
		/////////////////////////////////////////////////
		// If the slot filler is a vector (of frames), //
		// submit each element as an extraction prompt //
		/////////////////////////////////////////////////
		if (slotFiller instanceof Vector) {
			ListIterator li= ((Vector)slotFiller).listIterator();
			OrderedHashtable ht= new OrderedHashtable();
			while (li.hasNext()) {
				Frame element= null;
				try {
					element= (Frame)li.next();
					CCUS_FMA.indent();
					System.out.println("Submitting " +
					element + " as an extraction prompt");

					mergeHashtables(ht, ExtractionEngine.
					                CCUSextract(element));
					_numberOfHLCsExtracted0+=
					  ExtractionEngine.
					  _numberOfHLCsExtracted;
				}
				catch (ClassCastException cce) {
					System.err.println("Warning: HLC_CCUS"
					+ ".submitExtractionPrompts: "
					+ slotFiller + " contains non-frame "
					+ "elements!");
				}
			}
			return ht;
		} else
		/////////////////////////////////////////////////////////
		// If the slot filler is a non-frame/vector value,     //
		// submit the containing frame as an extraction prompt //
		/////////////////////////////////////////////////////////
		{
			CCUS_FMA.indent();
			System.out.println("Submitting " + frame
			+ " as an extraction prompt");

			OrderedHashtable returnValue= ExtractionEngine.
			                              CCUSextract(frame);
			_numberOfHLCsExtracted0= ExtractionEngine.
			                         _numberOfHLCsExtracted;			
			return returnValue;
		}
	}

	/********************************************************************/
	// HLC_CCUS.submitExtractionPrompts : (Frame x Frame x Slot)
	//                                    -> OrderedHashtable[2]
	/*********************************************************************
	** Considers two corresponding frames together for submitting
	** extraction prompts.
	** @param frame1 A frame from the first model to extract HLCs from.
	** @param frame2 A frame from the second model to extract HLCs from.
	** @param slot The slot for which a difference was found.
	** @return A pair of hashtables of the HLCs extracted (each key is a
	**         frame class, and each value is a vector of the HLCs of that
	**         frame class).
	*********************************************************************/

	private static OrderedHashtable[] submitExtractionPrompts
	                               (Frame frame1, Frame frame2, Slot slot)
	{
		Object slotFiller1= frame1.getSlotFiller(slot);
		Object slotFiller2= frame2.getSlotFiller(slot);

		//////////////////////////////////////////
		// If the slot fillers are both frames, //
		// submit them as extraction prompts    //
		//////////////////////////////////////////
		if (slotFiller1 instanceof Frame
		 && slotFiller2 instanceof Frame)
		{
			OrderedHashtable[] returnValue
			= new OrderedHashtable[2];

			CCUS_FMA.indent();
			System.out.println("Submitting " + slotFiller1
			+ " as an extraction prompt");
			returnValue[0]= ExtractionEngine.
			                CCUSextract((Frame)slotFiller1);
			_numberOfHLCsExtracted1+= ExtractionEngine.
			                          _numberOfHLCsExtracted;

			CCUS_FMA.indent();
			System.out.println("Submitting " + slotFiller2
			+ " as an extraction prompt");
			returnValue[1]= ExtractionEngine.
			                CCUSextract((Frame)slotFiller2);
			_numberOfHLCsExtracted2+= ExtractionEngine.
			                          _numberOfHLCsExtracted;

			return returnValue;
			
		} else
		///////////////////////////////////////////////////////
		// If the slot fillers are both vectors (of frames), //
		// submit each element from the both of them         //
		// as an extraction prompt                           //
		///////////////////////////////////////////////////////
		if (slotFiller1 instanceof Vector
		 && slotFiller2 instanceof Vector)
		{
			ListIterator li= ((Vector)slotFiller1).listIterator();
			OrderedHashtable ht1= new OrderedHashtable();
			while (li.hasNext()) {
				Frame element= null;
				try {
					element= (Frame)li.next();
					CCUS_FMA.indent();
					System.out.println("Submitting " +
					element + " as an extraction prompt");

					mergeHashtables(ht1, ExtractionEngine.
					                CCUSextract(element));
					_numberOfHLCsExtracted1+=
					  ExtractionEngine.
					  _numberOfHLCsExtracted;
				}
				catch (ClassCastException cce) {
					System.err.println("Warning: HLC_CCUS"
					+ ".submitExtractionPrompts: "
					+ slotFiller1 + " contains non-frame "
					+ "elements!");
				}
			}
			li= ((Vector)slotFiller2).listIterator();
			OrderedHashtable ht2= new OrderedHashtable();
			while (li.hasNext()) {
				Frame element= null;
				try {
					element= (Frame)li.next();
					CCUS_FMA.indent();
					System.out.println("Submitting " +
					element + " as an extraction prompt");

					mergeHashtables(ht2, ExtractionEngine.
					                CCUSextract(element));
					_numberOfHLCsExtracted2+=
					  ExtractionEngine.
					  _numberOfHLCsExtracted;
				}
				catch (ClassCastException cce) {
					System.err.println("Warning: HLC_CCUS"
					+ ".submitExtractionPrompts: "
					+ slotFiller2 + " contains non-frame "
					+ "elements!");
				}
			}
			OrderedHashtable[] returnValue
			= new OrderedHashtable[2];
			returnValue[0]= ht1; returnValue[1]= ht2;
			return returnValue;
		} else
		//////////////////////////////////////////////////////////
		// If either of the slot fillers is a non-frame/vector, //
		// submit the containing frame as an extraction prompt  //
		//////////////////////////////////////////////////////////
		{
			OrderedHashtable[] returnValue
			= new OrderedHashtable[2];

			CCUS_FMA.indent();
			System.out.println("Submitting " + frame1
			+ " as an extraction prompt");
			returnValue[0]= ExtractionEngine.CCUSextract(frame1);
			_numberOfHLCsExtracted1+= ExtractionEngine.
			                          _numberOfHLCsExtracted;			

			CCUS_FMA.indent();
			System.out.println("Submitting " + frame2
			+ " as an extraction prompt");
			returnValue[1]= ExtractionEngine.CCUSextract(frame2);
			_numberOfHLCsExtracted2+= ExtractionEngine.
			                          _numberOfHLCsExtracted;			

			return returnValue;
		}
	}

	/********************************************************************/
	// HLC_CCUS.mergeHashtables : (OrderedHashtable x OrderedHashtable)
	//                            -> OrderedHashtable
	/*********************************************************************
	** Merges two hash tables of vectors into a single hashtable where
	** corresponding vectors are concatenated. Works by copying the
	** contents of the second hashtable into the first, so when merging a
	** series of hashtables, this method should be called with the big
	** running hashtable as the first argument.
	** @param hashtable1 The first hashtable - the big, running hashtable,
	**                   which will have the contents of the second copied
	**                   into it.
	** @param hashtable1 The second hashtable.
	** @return The first hashtable, with the contents of the second copied
	**         into it.
	*********************************************************************/

	private static OrderedHashtable mergeHashtables
	            (OrderedHashtable hashtable1, OrderedHashtable hashtable2)
	{
		ListIterator ht2= hashtable2.keyListIterator();
		while (ht2.hasNext()) {
			Object key= ht2.next();
			Vector v1= (Vector)hashtable1.get(key);
			if (v1 == null) {
				v1= new Vector();
				hashtable1.put(key,v1);
			}
			Vector v2= (Vector)hashtable2.get(key);

			v1.addAll(v2);
		}
		return hashtable1;
	}
}
