      ////////////////////////////////////////
     // THE CONSTRAINT-UNDERSTANDING SYSTEM //
    //  (Version 2)                        ///
   // Written by Andrew Broad             ////
  //                                     /////
 // JavaCUS.java (2001:09:20 15:30)     //////
//////////////////////////////////////////////
// Methods of JavaCUS:                 ///////
// - static getExtractionCaseLibrary   ///////
// - static getNavigationCaseLibrary   ///////
// - static getPrettyPrintCaseLibrary  ///////
// - static writeClass                 ///////
// - static readClass                  ///////
// - static main                       //////
// - static usage                      /////
// - static processArguments           ////
// - static processFlags               ///
// - time                              //
////////////////////////////////////////

import java.util.*;
import java.io.*;

import jex.*;

/*****************************************************************************
** The class `JavaCUS' is the top-level class of the Java Constraint-
** Understanding System. It is not meant to be instantiated.
*****************************************************************************/

public class JavaCUS {
	public static boolean _generateHTML= false;
	public static boolean _onlyInputModelFramesInHTML= false;
	public static boolean _showTimes= false;

	private static CaseLibrary _extractionCaseLibrary= new CaseLibrary();
	private static CaseLibrary _navigationCaseLibrary= new CaseLibrary();
	private static CaseLibrary _ppCaseLibrary= new CaseLibrary();

	private static String _javaFilename= null;
	private static String _caseLibraryFilename
	       = "/home/broada/jws/PhD_CUS/CASES/library.cas";
	private static boolean _parseFirst= false;

	/********************************************************************/
	// static JavaCUS.getExtractionCaseLibrary : void -> CaseLibrary
	/*********************************************************************
	** Returns a reference to the extraction case library of the Java
	** CUS.
	** @return The extraction case library.
	*********************************************************************/

	public static CaseLibrary getExtractionCaseLibrary () {
		return _extractionCaseLibrary;
	}

	/********************************************************************/
	// static JavaCUS.getNavigationCaseLibrary : void -> CaseLibrary
	/*********************************************************************
	** Returns a reference to the navigation case library of the Java
	** CUS.
	** @return The navigation case library.
	*********************************************************************/

	public static CaseLibrary getNavigationCaseLibrary () {
		return _navigationCaseLibrary;
	}

	/********************************************************************/
	// static JavaCUS.getPrettyPrintCaseLibrary : void -> CaseLibrary
	/*********************************************************************
	** Returns a reference to the pretty-print case library of the Java
	** CUS.
	** @return The pretty-print case library.
	*********************************************************************/

	public static CaseLibrary getPrettyPrintCaseLibrary () {
		return _ppCaseLibrary;
	}

	/********************************************************************/
	// static JavaCUS.writeClass : ObjectOutput -> void
	/*********************************************************************
	** Serialises the static data members (class variables) of JavaCUS.
	** @param oo An output stream to save the class data to.
	** @exception IOException If an I/O exception occurs.
	*********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_extractionCaseLibrary);
		oo.writeObject(_navigationCaseLibrary);
		oo.writeObject(_ppCaseLibrary);
	}

	/********************************************************************/
	// static JavaCUS.readClass : ObjectInputStream -> void
	/*********************************************************************
	** Deserialises the static data members (class variables) of JavaCUS.
	** @param oi An input stream to save the class data to.
	** @exception IOException If an I/O exception occurs.
	*********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_extractionCaseLibrary= (CaseLibrary)oi.readObject();
			_navigationCaseLibrary= (CaseLibrary)oi.readObject();
			_ppCaseLibrary= (CaseLibrary)oi.readObject();
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the st"
			+ "upid compiler is just too dumb to realise this!");
		}
	}

	/********************************************************************/
	// JavaCUS.main : String -> void
	/*********************************************************************
	** The main entry point to the Java CUS.
	** @param argv The vector of command line arguments. There should be
	**  two of these: the filename of an Java compilation unit (.java) to
	**  parse, and a case library (.cas). If the latter is not
	**  specified, it defaults to "CASES/library.cas". There can also be
	**  command-line flags:
	**  -p causes the HLCs in the summary to be pretty-printed,
	**  -g causes HTML files for frames to be generated without asking,
	**  -i causes only input model frames to be included in the HTML,
	**  -s causes the intermediate output of the CUS to be suppressed,
	**  -t prints timestamps at the major phase boundaries.
	*********************************************************************/

	public static void main (String argv[]) {
		try {
			System.setProperty("java.compiler","none");
			System.setProperty("jex.licenseLocation",
			                   "27000@manatee");

			processArguments(argv);

			Frame._markFramesAsInputModel= true;

			Frame inputFrameNetwork= null;

			FileInputStream fis;

			if (_parseFirst) {
				System.out.println(time() + "Parsing Java com"
				+ "pilation unit in \"" +_javaFilename+ "\"");
				JavaParser parser= new JavaParser
				                   (fis= new FileInputStream
				                    (_javaFilename));
				inputFrameNetwork= parser.CompilationUnit();
				fis.close();
				System.out.println("Done parsing Java compila"
				+ "tion unit.\n");

				System.out.println(time() + "Loading case lib"
				+"rary from \""+_caseLibraryFilename+"\"...");
				CaseLibrary.load(_caseLibraryFilename);
				System.out.println
				           ("Done loading case library.\n");
			} else {
				System.out.println(time() + "Loading case lib"
				+"rary from \""+_caseLibraryFilename+"\"...");
				CaseLibrary.load(_caseLibraryFilename);
				System.out.println
				           ("Done loading case library.\n");

				System.out.println(time() + "Parsing Java com"
				+ "pilation unit in \"" +_javaFilename+ "\"");
				JavaParser parser= new JavaParser
				                   (fis= new FileInputStream
				                    (_javaFilename));
				inputFrameNetwork= parser.CompilationUnit();
				fis.close();
				System.out.println("Done parsing Java compila"
				+ "tion unit.\n");
			}

			System.out.println(time() + "Extracting higher-level "
			+ "knowledge...\n");
			ExtractionEngine.extraction(inputFrameNetwork,
			                            _extractionCaseLibrary,
			                            _navigationCaseLibrary,
			                            _ppCaseLibrary);
		} catch (Throwable t) {
			System.err.println(
				"----------------------------------------" +
				"----------------------------------------");
			System.err.println(time() + "A Throwable of type "
			+ t.getClass() + " has been caught in "
			+ "JavaCUS.main!\n\ttoString == " + t + "\n");
			System.err.println("Its Message is:-\n"
			                  + t.getMessage() + "\n");
			System.err.println("Its LocalizedMessage is:-\n"
			+ t.getLocalizedMessage());
			System.err.println("Printing stack trace:-\n");
			t.printStackTrace();
		}
		System.out.println(
			"----------------------------------------" +
			"----------------------------------------");
		if (! _generateHTML) {
			System.out.print(time() + "Generate HTML for the fram"
			+ "es (y/n)? ");
			BufferedReader input= new BufferedReader
			                   (new InputStreamReader(System.in));
			String generateHTML= null;
			try {
				generateHTML= input.readLine();
			} catch (IOException ioe) {
				System.err.println("JavaCUS.main: IOException"
				+ " thrown by BufferedReader.readLine()");
				System.exit(1);
			}
			System.out.println();
			_generateHTML= generateHTML.
			               regionMatches(true,0,"y",0,1);
		}
		if (_generateHTML) {
			System.out.println("\n" + time() + "Generating HTML "
			+ "in /usr/local/tmp/users/broada/FRAMES/...");
			Frame.generateFIL(true, _onlyInputModelFramesInHTML,
			                  null);
		}
		if (_showTimes) {
			System.out.println("\n" + time() + "Finished.");
		}
	}

	/********************************************************************/
	// static JavaCUS.usage : void -> void
	/*********************************************************************
	** Print the usage report when the command-line arguments are invalid,
	** and exit the system.
	*********************************************************************/

	private static void usage () {
		System.err.println("Usage: Two arguments: an Java compilation"
		+ " unit (.java file) and a case library (.cas file). The lat"
		+ "ter is optional - if omitted, the default is \"CASES/libra"
		+ "ry.cas\"");
		System.err.println("    Flags:-");
		System.err.println("-p: Pretty-print HLCs in summary");
		System.err.println("-g: Generate HTML without asking");
		System.err.println("-i: Include only input model frames in "
		+ "HTML");
		System.err.println("-s: Suppress intermediate output");
		System.err.println("-t: Show times at phase boundaries");
		System.exit(1);
	}

	/********************************************************************/
	// static JavaCUS.processArguments : String[] -> void
	/*********************************************************************
	** A subroutine to process the command-line arguments to the CCUS.
	** @param argv The vector of command-line arguments. There should be
	**  three of these: the filenames of two EXPRESS models (.exp) to
	**  parse using jex, and a case library (.cas). If the latter is not
	**  specified, it defaults to "CASES/library.cas".
	*********************************************************************/

	private static void processArguments (String argv[]) {
		for (int i= 0; i < argv.length; i++) {
			if (argv[i].endsWith(".java")) {
				_javaFilename= argv[i];
				_parseFirst= false;
			}
			if (argv[i].endsWith(".cas")) {
				_caseLibraryFilename= argv[i];
				_parseFirst= true;
			}
			if (argv[i].startsWith("-")) {
				processFlags(argv[i]);
			}
		}

		if (_javaFilename == null) {
			usage();
		}
	}

	/********************************************************************/
	// static JavaCUS.processFlags : String[] -> void
	/*********************************************************************
	** A subroutine to process the flags in a command-line argument
	** starting with `-'. This method assumes that it is only called with
	** strings starting with `-', and ignores the first character.
	** @param arg A `flags' command-line argument (i.e. `-' followed by
	**            letters).
	*********************************************************************/

	private static void processFlags (String arg) {
		char[] flags= arg.toCharArray();

		for (int i= 1; i < flags.length; i++) {
			switch (flags[i]) {
				case 'p':
					ExtractionEngine.
					_prettyPrintHLCs= true;
					break;
				case 's':
					ExtractionEngine.
					_showIntermediateOutput= false;
					break;
				case 'g':
					_generateHTML= true;
					break;
				case 'i':
					_onlyInputModelFramesInHTML= true;
					break;
				case 't':
					_showTimes= true;
					break;
				default:
					System.err.println("Warning: -"
					+ flags[i] + " not recognised.");
			}
		}
	}

	/********************************************************************/
	// static JavaCUS.time : void -> String
	/*********************************************************************
	** If _showTimes is true, return the current time as a string of the
	** form "(HH:MM:SS.SSS) ", otherwise return the empty string.
	** @return The time string.
	*********************************************************************/

	private static String time () {
		String time= "";
		if (_showTimes) {
			GregorianCalendar gregorc= new GregorianCalendar();
			//////////////
			// hours... //
			//////////////
			time= "(";
			int hours= gregorc.get(GregorianCalendar.HOUR_OF_DAY);
			if (hours < 10) time+= "0";
			time+= hours + ":";
			////////////////
			// minutes... //
			////////////////
			int minutes= gregorc.get(GregorianCalendar.MINUTE);
			if (minutes < 10) time+= "0";
			time+= minutes + ":";
			////////////////
			// seconds... //
			////////////////
			int seconds= gregorc.get(GregorianCalendar.SECOND);
			if (seconds < 10) time+= "0";
			time+= seconds + ".";
			/////////////////////
			// milliseconds... //
			/////////////////////
			int ms= gregorc.get(GregorianCalendar.MILLISECOND);
			if (ms < 100) {
				time+= "0";
				if (ms < 10) time+= "0";
			}
			time+= ms + ") ";
		}
		return time;
	}
}
