      ////////////////////////////////////////
     // THE CONSTRAINT-UNDERSTANDING SYSTEM //
    //  (Version 2)                        ///
   // Written by Andrew Broad             ////
  //                                     /////
 // JexToFrames.java (2002:07:29 16:23) //////
//////////////////////////////////////////////
// Methods of JexToFrames:             ///////
// - getCachedFrame                    ///////
// - convert(Model,boolean)            ///////
// - convert(JexObject)                ///////
// - convertListIterator               ///////
// - convertRef                        ///////
// - convertSimpleType                 ///////
// - convertInverseAttribute           ///////
// - convertInverseQualifiedAttribute  ///////
// - convertEnumerationType            ///////
// - convertSelectType                 ///////
// - convertSupertypeExpression        ///////
// - convertSupertypeFactor            ///////
// - convertExpression                 ///////
// - convertSimpleExpression           ///////
// - convertTerm                       ///////
// - convertFactor                     ///////
// - convertQualifiedFactor            ///////
// - convertQualifier                  ///////
// - convertUserDefinedFunctionCall    ///////
// - convertInterval                   ///////
// - convertUnaryOperationFactor       ///////
// - convertSetType                    ///////
// - convertGeneralSetType             ///////
// - convertBagType                    ///////
// - convertGeneralBagType             ///////
// - convertListType                   //////
// - convertGeneralListType            /////
// - isTypeDiscrete                    ////
// - writeClass                        ///
// - readClass                         //
////////////////////////////////////////

import java.util.*;
import java.io.*;

import jex.*;

/*****************************************************************************
** The class `JexToFrames' is responsible for converting Jex objects to
** frames. It is not meant to be instantiated by the client.
*****************************************************************************/

public class JexToFrames {
	private static Hashtable/*JexObject->Frame*/ _cache= new Hashtable();
	private static Slot _slotBeingFilled;

	/********************************************************************/
	// static JexToFrames.getCachedFrame : Object -> Frame
	/*********************************************************************
	** Returns the frame associated with a given Jex object. Exits with an
	** internal error if it is not in the cache.
	** @param jObject The Jex object.
	** @return The frame associated with the Jex object.
	*********************************************************************/

	public static Frame getCachedFrame (Object jObject) {
		Frame frame= (Frame)_cache.get(jObject);
		if (frame == null) {
			System.err.println("Error: " + jObject + " not in the"
			+ " JexToFrames cache!");
			System.exit(1);
		}
		return frame;
	}

	/********************************************************************/
	// static JexToFrames.convert : Model x boolean -> Frame
	/*********************************************************************
	** Converts the objects in the given Jex EXPRESS model to frames.
	** @param model The EXPRESS model.
	** @param createFrameClasses If true, will also create a FrameClass
	**  for each entity, a Slot for each attribute and a DefinedType for
	**  each defined type in the model.
	** @return The top level frame of the model (i.e. the "model_frame").
	*********************************************************************/

	public static Frame convert (Model model,boolean createFrameClasses) {
		if (createFrameClasses) {
			ListIterator schemata= model.getSchemas();
			while (schemata.hasNext()) {
				Schema schema= (Schema)schemata.next();

				FrameClass.createFrameClasses(schema);
				DefinedType.createDefinedTypes(schema);
			}
		}
		return (Frame)convert(model);
	}

	/********************************************************************/
	// static JexToFrames.convert : JexObject -> Object
	/*********************************************************************
	** Converts a given Java object to a frame object. Works by looking up
	** a frame template appropriate for the class of the object, and using
	** that to do the conversion in a data-driven manner. ListIterator
	** objects are converted to FIL vectors. Other objects for which no
	** frame template exists are returned as they are.
	** @param jObject The object to be converted to a frame (or whatever).
	** @return The converted object.
	*********************************************************************/

	private static Object convert (Object jObject) {
		if (jObject == null) {
			return null;
		}

		//////////////////////////////////////////////////////////////
		// If a frame has already been created for the Java object, //
		// just return the reference to that frame.                 //
		//////////////////////////////////////////////////////////////
		Frame frame= (Frame)_cache.get(jObject);
		if (frame != null)
			return frame;

		///////////////////////////////////////////////////////
		// Hardwired conversions for some specific Jex types //
		///////////////////////////////////////////////////////

		if (jObject instanceof ListIterator) {
			return convertListIterator((ListIterator)jObject);
		}

		if (jObject instanceof String) {
			return new FILString((String)jObject);
		}

		if (jObject instanceof Ref) {
			return convertRef((Ref)jObject);
		}

		if (jObject instanceof NestedExpression) {
			try {
				return convert(((NestedExpression)jObject).
				               getNestedExpression());
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convert("
				+ jObject + "): UnsetValueException for "
				+ "jObject.getNestedExpression()");
				System.exit(1); return null;
			}
		}

		if (jObject instanceof Bucket) {
			try {
				if (((Bucket)jObject).getActualParameters().
				    hasNext())
				{
					return convertUserDefinedFunctionCall
					       (convertRef(((Bucket)jObject).
					        getRef()),
				      	       ((Bucket)jObject).
				      	        getActualParameters());
				}
				if (((Bucket)jObject).hasQualifier()) {
					return convertQualifier(convertRef
				      	       (((Bucket)jObject).getRef()),
				      	    ((Bucket)jObject).getQualifier());
				}
				return convertRef
				       (((Bucket)jObject).getRef());
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convert"
				+ "(Object): UnsetValueException for "
				+ "((Bucket)jObject).getRef()");
				System.exit(1);
			}
		}

		if (jObject instanceof SimpleExpression) {
			return convertSimpleExpression
			       ((SimpleExpression)jObject);
		}

		if (jObject instanceof Expression) {
			return convertExpression((Expression)jObject);
		}

		if (jObject instanceof SimpleTypes) {
			return convertSimpleType((SimpleTypes)jObject);
		}

		if (jObject instanceof GenericType) {
			return new DefinedTypeInstance(DefinedType.lookup
			     ("simple_type"), new EnumSymbol("generic_type"));
		}

		if (jObject instanceof InverseAttribute) {
			return convertInverseAttribute
			       ((InverseAttribute)jObject);
		}

		if (jObject instanceof InverseQualifiedAttribute) {
			return convertInverseQualifiedAttribute
			       ((InverseQualifiedAttribute)jObject);
		}

		if (jObject instanceof SetType) {
			return convertSetType((SetType)jObject);
		}

		if (jObject instanceof GeneralSetType) {
			return convertGeneralSetType((GeneralSetType)jObject);
		}

		if (jObject instanceof BagType) {
			return convertBagType((BagType)jObject);
		}

		if (jObject instanceof GeneralBagType) {
			return convertGeneralBagType((GeneralBagType)jObject);
		}

		if (jObject instanceof ListType) {
			return convertListType((ListType)jObject);
		}

		if (jObject instanceof GeneralListType) {
			return convertGeneralListType
			       ((GeneralListType)jObject);
		}

		if (jObject instanceof EnumerationType) {
			return convertEnumerationType
			       ((EnumerationType)jObject);
		}

		if (jObject instanceof SelectType) {
			return convertSelectType((SelectType)jObject);
		}

		if (jObject instanceof SupertypeExpression) {
			return convertSupertypeExpression
			       ((SupertypeExpression)jObject);
		}

		if (jObject instanceof QualifiedFactor) {
			return convertQualifiedFactor
			       ((QualifiedFactor)jObject);
		}

		if (jObject instanceof Interval) {
			return convertInterval((Interval)jObject);
		}

		if (jObject instanceof UnaryOperationFactor) {
			return convertUnaryOperationFactor
			       ((UnaryOperationFactor)jObject);
		}

		if (jObject instanceof Boolean) {
			if (((Boolean)jObject).booleanValue())
				return Logical.TRUE;
			else	return Logical.FALSE;
		}

		if (jObject instanceof LogicalLiteral) {
			frame=new Frame(FrameClass.lookup("logical_literal"));
			try {
				Logical l= null;
				if (((LogicalLiteral)jObject).getValue()
				 == LogicalLiteral.TRUE)
					l= Logical.TRUE;
				if (((LogicalLiteral)jObject).getValue()
				 == LogicalLiteral.FALSE)
					l= Logical.FALSE;
				if (((LogicalLiteral)jObject).getValue()
				 == LogicalLiteral.UNKNOWN)
					l= Logical.UNKNOWN;
				if (l == null) {
					System.err.println("JexToFrames.conve"
					+ "rt: failed to convert LogicalLiter"
					+ "al " + jObject);
					System.exit(1);
				}					
				frame.ssf("literal_value",l);
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convert: "
				+ "UnsetValueException for BinaryLiteral."
				+ "getValue()");
				System.exit(1);
			}
			frame.ssf("tipe", new DefinedTypeInstance
			                  (DefinedType.lookup("simple_type"),
			                   new EnumSymbol("logical_type")));
			frame.ssf("is_variable", Logical.FALSE);
			return frame;
		}

		if (jObject instanceof BinaryLiteral) {
			frame= new Frame(FrameClass.lookup("binary_literal"));
			try {
				frame.ssf("literal_value",
				          new Binary(((BinaryLiteral)jObject).
				                     getValue()));
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convert: "
				+ "UnsetValueException for BinaryLiteral."
				+ "getValue()");
				System.exit(1);
			}
			frame.ssf("tipe", new DefinedTypeInstance
			                  (DefinedType.lookup("simple_type"),
			                   new EnumSymbol("binary_type")));
			frame.ssf("is_variable", Logical.FALSE);
			return frame;
		}

		if (jObject instanceof RulePopulationStmtVariable) {
			try {
				return convertRef(((
				RulePopulationStmtVariable)jObject).getRef());
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convert"
				+ "(Object): UnsetValueException for ((Rule"
				+ "PopulationStmtVariable)jObject).getRef()");
				System.exit(1);
			}
		}

		////////////////////////////////////////////////////////////
		// Otherwise, look up a frame template for converting the //
		// Java object to a frame.                                //
		////////////////////////////////////////////////////////////
		FrameTemplate template= FrameTemplate.find
		                                      (jObject.getClass());

		////////////////////////////////////////
		// If no frame template exists,       //
		// just return the Java object as is. //
		////////////////////////////////////////
		if (template == null) {
			if (jObject.toString().indexOf('@') != -1) {
				System.err.println("Warning: " + jObject
				+ " not converted");
			}
			return jObject;
		}

		////////////////////////////////////////////////////////
		// Create the frame for the Java object and cache the //
		// association for future reference.                  //
		////////////////////////////////////////////////////////
		frame= new Frame(template.getFrameClass());
		_cache.put(jObject,frame);

		///////////////////////////////////////////////////////////
		// Use the frame template to fill the frame's slots from //
		// the properties of the Java object.                    //
		///////////////////////////////////////////////////////////
		ListIterator filledSlots= template.getSlots();
		while (filledSlots.hasNext()) {
			Slot slot= (Slot)filledSlots.next();
			_slotBeingFilled= slot;
			Object filler= template.getSlotFiller(slot);
			if (filler instanceof InvocationList) {
				Object result= ((InvocationList)filler).
				               invoke(jObject);
				filler= JexToFrames.convert(result);
			}
			frame.setSlotFiller(slot,filler);
		}
		return frame;
	}

	/********************************************************************/
	// static JexToFrames.convertListIterator : ListIterator -> Vector
	/*********************************************************************
	** Converts a ListIterator of Java (Jex) objects to a Vector of
	** FIL objects - either an OrderedVector or an UnorderedVector as
	** appropriate. Each element of the list is processed recursively to
	** convert Jex objects to frames.
	** @param listIterator The list iterator to be converted.
	** @return The vector into which the listIterator has been converted.
	*********************************************************************/

	private static Vector convertListIterator (ListIterator listIterator){
		if (! listIterator.hasNext()
		   && _slotBeingFilled.convertEmptyToNull())
			return null;

		Vector vector= null;
		if (_slotBeingFilled.isOrderedAggType())
			vector= new OrderedVector();
		else if (_slotBeingFilled.isUnorderedAggType())
			vector= new UnorderedVector();
		else {
			System.err.println(_slotBeingFilled + " is of an aggr"
			+ "egation type which is not a SET, BAG, LIST or ARRA"
			+ "Y!");
			//System.exit(1);
		}

		while (listIterator.hasNext()) {
			vector.addElement(JexToFrames.convert
			                  (listIterator.next()));
		}
		return vector;
	}

	/********************************************************************/
	// static JexToFrames.convertRef : Ref -> Object
	/*********************************************************************
	** Converts a Ref to an object in the frame system. The Ref is
	** resolved, and the referent is converted recursively. If the Ref
	** cannot be resolved, a value of type "unresolved_reference" is
	** returned.
	** @param ref The reference to be converted.
	** @return The object into which the ref has been converted.
	*********************************************************************/

	private static Object convertRef (Ref ref) {
		try {
			return convert(ref.getReferencedObject());
		}
		catch (UnsetValueException uve) {
			String refString= "";
			try {
				refString= ref.getId();
			}
			catch (UnsetValueException uve2) {
				System.err.println("JexToFrames.convertRef: "
				+ "nested UnsetValueException!");
				System.exit(1);
			}
			System.err.println("Warning: " + ref
			  + " \"" + refString + "\""
			  + " could not be resolved.");
			return new DefinedTypeInstance
			  (DefinedType.lookup("unresolved_reference"),
			  new FILString(refString));
		}
	}

	/********************************************************************/
	// static JexToFrames.convertSimpleType :
	//                    SimpleTypes -> DefinedTypeInstance
	/*********************************************************************
	** Converts a SimpleType to an defined type instance.
	** @param simpleType The simple type to be converted.
	** @return The defined type instance into which the simpleType has
	**         been converted.
	*********************************************************************/

	private static DefinedTypeInstance convertSimpleType
	                                   (SimpleTypes simpleType)
	{
		String enum= "";
		if (simpleType instanceof BinaryType)
			enum= "binary_type";
		else if (simpleType instanceof BooleanType)
			enum= "boolean_type";
		else if (simpleType instanceof IntegerType)
			enum= "integer_type";
		else if (simpleType instanceof LogicalType)
			enum= "logical_type";
		else if (simpleType instanceof NumberType)
			enum= "number_type";
		else if (simpleType instanceof RealType)
			enum= "real_type";
		else if (simpleType instanceof StringType)
			enum= "string_type";
		return new DefinedTypeInstance(DefinedType.lookup
		       ("simple_type"), new EnumSymbol(enum));
	}

	/********************************************************************/
	// static JexToFrames.convertInverseAttribute :
	//                    InverseAttribute -> Object
	/*********************************************************************
	** Converts a Jex InverseAttribute object to an "inverse_attribute"
	** frame.
	** @param inverseAttribute The inverse attribute to be converted.
	** @return The frame into which the inverse attribute has been
	**         converted.
	*********************************************************************/

	private static Frame convertInverseAttribute
	                     (InverseAttribute inverseAttribute)
	{
		Frame frame=new Frame(FrameClass.lookup("inverse_attribute"));
		String forAttributeId= "NOT_SET_YET";
		try {
			frame.ssf("name",
			          new FILString(inverseAttribute.getId()));

			/////////////////////////////////////////////////
			// An inverse attribute can be either a SET or //
			// a BAG, with an optional bound spec.         //
			/////////////////////////////////////////////////

			Object type0=convert(inverseAttribute.getEntityRef());

			if (inverseAttribute.isSet()) {
				Frame type1= new Frame
				             (FrameClass.lookup("set_type"));
				type1.ssf("element_type", type0);
				if (inverseAttribute.hasBoundSpec()) {
					type1.ssf("lower_bound", convert
					  (inverseAttribute.getBoundSpec().
					    getLowerBound()));
					type1.ssf("upper_bound", convert
					  (inverseAttribute.getBoundSpec().
					    getUpperBound()));
				}
				type0= type1;
			}
			if (inverseAttribute.isBag()) {
				Frame type1= new Frame
				             (FrameClass.lookup("bag_type"));
				type1.ssf("element_type", type0);
				if (inverseAttribute.hasBoundSpec()) {
					type1.ssf("lower_bound", convert
					  (inverseAttribute.getBoundSpec().
					    getLowerBound()));
					type1.ssf("upper_bound", convert
					  (inverseAttribute.getBoundSpec().
					    getUpperBound()));
				}
				type0= type1;
			}

			frame.ssf("tipe", type0);

			////////////////////////////////////////////////////
			// Special reference resolution for FOR attribute //
			////////////////////////////////////////////////////
			Ref ref= inverseAttribute.getAttributeRef();
			ExplicitAttribute forAttribute= null;
			try {
				forAttribute= (ExplicitAttribute)ref.
				              getReferencedObject();
			}
			catch (Exception e) {
				System.err.println("Jex failed to resolve the"
				+ " FOR attribute of "
				+ inverseAttribute.getId() + " (threw " + e
				+ ") - using specialist FOR reference resolut"
				+ "ion procedure");

				forAttributeId= ref.getId();
				Entity entity= (Entity)inverseAttribute.
				                       getEntityRef().
				                       getReferencedObject();
				try {
					forAttribute= entity.
					findExplicitAttribute(forAttributeId);
				} catch (NullParameterException npe) {
					System.err.println("JexToFrames.conve"
					+ "rtInverseAttribute: NullParameterE"
					+ "xception thrown by Entity.findExpl"
					+ "icitAttribute");
					System.exit(1);
				}
			}
			frame.ssf("for_attribute", convert(forAttribute));
		} catch (UnsetValueException uve) {
			System.err.println("JexToFrames.convertInverseAttribu"
			+ "te: UnsetValueException");
			frame.ssf("for_attribute", new DefinedTypeInstance
			  (DefinedType.lookup("unresolved_reference"),
			  new FILString(forAttributeId)));
		}
		if (_cache.get(inverseAttribute) != null) {
			System.err.println("JexToFrames.convertInverseAttribu"
			+ "te: Warning: (" + inverseAttribute + ", " + frame
			+ " already in JexToFrames._cache!");
		}
		_cache.put(inverseAttribute,frame);
		return frame;
	}

	/********************************************************************/
	// static JexToFrames.convertInverseQualifiedAttribute :
	//                    InverseQualifiedAttribute -> Object
	/*********************************************************************
	** Converts a Jex InverseQualifiedAttribute object to an
	** "inverse_attribute_redeclaration" frame.
	** @param inverseQualifiedAttribute The inverse attribute
	**                                  redeclarationto be converted.
	** @return The frame into which the inverse attribute redeclaration
	**         has been converted.
	*********************************************************************/

	private static Frame convertInverseQualifiedAttribute
	                     (InverseQualifiedAttribute iQA)
	{
		Frame frame= new Frame(FrameClass.lookup
		                       ("inverse_attribute_redeclaration"));
		String forAttributeId= "NOT_SET_YET";
		try {
			frame.ssf("attribute", convertRef
			     (iQA.getQualifiedAttribute().getAttributeRef()));
			frame.ssf("entity_originally_declared_in", convertRef
			     (iQA.getQualifiedAttribute().getEntityRef()));

			/////////////////////////////////////////////////
			// An inverse attribute can be either a SET or //
			// a BAG, with an optional bound spec.         //
			/////////////////////////////////////////////////

			Object type0= convert(iQA.getEntityRef());

			if (iQA.isSet()) {
				Frame type1= new Frame
				             (FrameClass.lookup("set_type"));
				type1.ssf("element_type", type0);
				if (iQA.hasBoundSpec()) {
					type1.ssf("lower_bound", convert
					(iQA.getBoundSpec().getLowerBound()));
					type1.ssf("upper_bound", convert
					(iQA.getBoundSpec().getUpperBound()));
				}
				type0= type1;
			}
			if (iQA.isBag()) {
				Frame type1= new Frame
				             (FrameClass.lookup("bag_type"));
				type1.ssf("element_type", type0);
				if (iQA.hasBoundSpec()) {
					type1.ssf("lower_bound", convert
					(iQA.getBoundSpec().getLowerBound()));
					type1.ssf("upper_bound", convert
					(iQA.getBoundSpec().getUpperBound()));
				}
				type0= type1;
			}

			frame.ssf("tipe", type0);

			////////////////////////////////////////////////////
			// Special reference resolution for FOR attribute //
			////////////////////////////////////////////////////
			Ref ref= iQA.getAttributeRef();
			ExplicitAttribute forAttribute= null;
			try {
				forAttribute= (ExplicitAttribute)ref.
				              getReferencedObject();
			}
			catch (Exception e) {
				System.err.println("Jex failed to resolve the"
				+ " FOR attribute of " + iQA + " (threw " + e
				+ ") - using specialist FOR reference resolut"
				+ "ion procedure");

				forAttributeId= ref.getId();
				Entity entity= (Entity)iQA.getEntityRef().
				                       getReferencedObject();
				try {
					forAttribute= entity.
					findExplicitAttribute(forAttributeId);
				} catch (NullParameterException npe) {
					System.err.println("JexToFrames.conve"
					+ "rtInverseQualifiedAttribute: NullP"
					+ "arameterException thrown by Entity"
					+ ".findExplicitAttribute");
					System.exit(1);
				}
			}
			frame.ssf("for_attribute", convert(forAttribute));
		} catch (UnsetValueException uve) {
			System.err.println("JexToFrames.convertInverseQualifi"
			+ "edAttribute: UnsetValueException");
			frame.ssf("for_attribute", new DefinedTypeInstance
			  (DefinedType.lookup("unresolved_reference"),
			  new FILString(forAttributeId)));
		}
		if (_cache.get(iQA) != null) {
			System.err.println("JexToFrames.convertInverseQualifi"
			+ "edAttribute: Warning: (" + iQA + ", " + frame
			+ " already in JexToFrames._cache!");
		}
		_cache.put(iQA,frame);
		return frame;
	}

	/********************************************************************/
	// static JexToFrames.convertEnumerationType :
	//                    EnumerationType -> DefinedTypeInstance
	/*********************************************************************
	** Converts an EnumerationType to an defined type instance.
	** @param enumerationType The enumeration type to be converted.
	** @return The defined type instance into which the enumerationType
	**         has been converted.
	*********************************************************************/

	private static DefinedTypeInstance convertEnumerationType
	                                   (EnumerationType enumerationType)
	{
		OrderedVector enumerationItems= new OrderedVector();
		ListIterator li= enumerationType.getEnumerationItems();
		while (li.hasNext()) {
			try {
				Frame enumSymbol= new Frame
				  (FrameClass.lookup("enumeration_symbol"));
				enumSymbol.ssf("name", new FILString
				  (((EnumerationItem)li.next()).getId()));
				enumerationItems.addElement(enumSymbol);
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convertEnum"
				+ "erationType: UnsetValueException");
				System.exit(1);
			}
		}
		return new DefinedTypeInstance(DefinedType.lookup
		       ("enumeration_type"), enumerationItems);
	}

	/********************************************************************/
	// static JexToFrames.convertSelectType :
	//                    SelectType -> DefinedTypeInstance
	/*********************************************************************
	** Converts a SelectType to an defined type instance.
	** @param selectType The select type to be converted.
	** @return The defined type instance into which the selectType has
	**         been converted.
	*********************************************************************/

	private static DefinedTypeInstance convertSelectType
	                                   (SelectType selectType)
	{
		UnorderedVector namedTypes= new UnorderedVector();
		ListIterator li= selectType.getNamedTypes();
		while (li.hasNext()) {
			namedTypes.addElement(convertRef((Ref)li.next()));
		}
		return new DefinedTypeInstance(DefinedType.lookup
		       ("select_type"), namedTypes);
	}

	/********************************************************************/
	// static JexToFrames.convertSupertypeExpression :
	//                    SupertypeExpression -> Object
	/*********************************************************************
	** Converts a SupertypeExpression to an ANDOR frame if there's more
	** than one factor; otherwise, just pass the factor on to
	** convertSupertypeFactor.
	** @param supertypeExpression The supertype expression to be
	**                            converted.
	** @return The frame into which the supertypeExpression has been
	**         converted.
	*********************************************************************/

	private static Object convertSupertypeExpression
	                     (SupertypeExpression supertypeExpression)
	{
		ListIterator li= supertypeExpression.getSupertypeFactors();

		switch (supertypeExpression.countSupertypeFactors()) {
			case 0:
				return null;
			case 1:
				return convertSupertypeFactor
				((SupertypeFactor)li.next());
			default:
				Frame andor= new Frame(FrameClass.lookup
				      ("andor_supertype_expression"));
				UnorderedVector factors=
				                new UnorderedVector();
				while (li.hasNext()) {
					factors.addElement
					  (convertSupertypeFactor
					    ((SupertypeFactor)li.next()));
				}
				try {
					andor.setSlotFiller
					      ("factors", factors);
				} catch (InvalidSlot is) {
					System.err.println("No slot \"factors"
					+ "\" of frame class \"andor_supertyp"
					+ "e_expression\"!");
					System.exit(1);
				}
				return andor;
		}
	}

	/********************************************************************/
	// static JexToFrames.convertSupertypeFactor :
	//                    SupertypeFactor -> Object
	/*********************************************************************
	** Converts a SupertypeFactor to an AND frame if there's more than one
	** term; otherwise just pass the term on to be converted.
	** @param supertypeFactor The supertype factor to be converted.
	** @return The frame into which the supertypeFactor has been
	**         converted.
	*********************************************************************/

	private static Object convertSupertypeFactor
	                      (SupertypeFactor supertypeFactor)
	{
		ListIterator li= supertypeFactor.getSupertypeTerms();

		switch (supertypeFactor.countSupertypeTerms()) {
			case 0:
				return null;
			case 1:
				return convert(li.next());
				//return convertSupertypeTerm
				//((SupertypeTerm)li.next());
			default:
				Frame and= new Frame(FrameClass.lookup
				      ("and_supertype_expression"));
				UnorderedVector terms=
				                new UnorderedVector();
				while (li.hasNext()) {
					terms.addElement(convert(li.next()));
					//(convertSupertypeTerm
					  //((SupertypeTerm)li.next()));
				}
				try {
					and.setSlotFiller("terms", terms);
				} catch (InvalidSlot is) {
					System.err.println("No slot \"terms\""
					+ " of frame class \"and_supertype_ex"
					+ "pression\"!");
					System.exit(1);
				}
				return and;
		}
	}

	/********************************************************************/
	// static JexToFrames.convertExpression : Expression -> Object
	/*********************************************************************
	** Converts an Expression to frames.
	** @param expression The expression to be converted.
	** @return The object into which the expression has been converted.
	*********************************************************************/

	private static Object convertExpression (Expression expression)
	{
		//////////////////////////////////////////////////////////////
		// Just pass unary operations on to convertSimpleExpression //
		//////////////////////////////////////////////////////////////
		if (! expression.hasSimpleExpression2()) {
			try {
				return convertSimpleExpression
				       (expression.getSimpleExpression1());
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convertExpres"
				+ "sion: UnsetValueException for expression.g"
				+ "etSimpleExpression1()");
				System.exit(1); return null;
			}
		} else {
		//////////////////////////
		// Non-unary operations //
		//////////////////////////
			String fcName= null;
			String slot1Name= null;
			String slot2Name= null;
			RelationalOperatorExtended operator= null;
			try {
				operator= expression.getOperator();
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames."
				+ "convertExpression: UnsetValueException for"
				+ " expression.getOperator()");
				System.exit(1);
			}
			/////////////////////////////////////////////////
			// Determine the frame class and slots to fill //
			// for the particular operation                //
			/////////////////////////////////////////////////
			if (operator instanceof GT || operator instanceof GE)
			{
				fcName= "comparator";
				slot1Name= "greater";
				slot2Name= "lesser";
			}
			if (operator instanceof LT || operator instanceof LE)
			{
				fcName= "comparator";
				slot1Name= "lesser";
				slot2Name= "greater";
			}
			if (operator instanceof EQ) {
				fcName= "equation";
				slot1Name= "equal_expressions";
			}
			if (operator instanceof INST_EQ) {
				fcName= "instance_equation";
				slot1Name= "equal_instances";
			}
			if (operator instanceof NE) {
				fcName= "inequality";
				slot1Name= "first_expression";
				slot2Name= "second_expression";
			}
			if (operator instanceof INST_NE) {
				fcName= "instance_inequality";
				slot1Name= "first_instance";
				slot2Name= "second_instance";
			}
			if (operator instanceof IN) {
				fcName= "in_operation";
				slot1Name= "element";
				slot2Name= "set_frame";
			}
			if (operator instanceof LIKE) {
				fcName= "like_operation";
				slot1Name= "source_string";
				slot2Name= "target_string";
			}
			//////////////////////////////////////////////////
			// Create the frame and start setting the slots //
			//////////////////////////////////////////////////
			Frame frame= new Frame(FrameClass.lookup(fcName));
			try {
				frame.ssf("tipe",
				          convert(expression.getType()));
			} catch (IncompatibleTypeException ite) {
				System.err.println("JexToFrames.convertExpres"
				+ "sion: Jex threw an IncompatibleTypeExcepti"
				+ "on for expression.getType()");
				System.err.println(ite.getLocalizedMessage());
				System.err.println("This is that stupid Jex "
				+ "1.9.3 error, so let's continue...");
				frame.ssf("tipe", new FILString
				                            (ite.toString()));
			} catch (ClassCastException cce) {
				System.err.println("JexToFrames.convertExpres"
				+ "sion: Jex threw a ClassCastException for e"
				+ "xpression.getType()");
				System.err.println(cce.getLocalizedMessage());
				System.err.println("This is that stupid Jex "
				+ "1.9.3 error, so let's continue...");
				frame.ssf("tipe", new FILString
				                            (cce.toString()));
			} catch (JexException je) {
				System.err.println("JexToFrames.convertExpres"
				+ "sion: " + je.getClass() + " for expression"
				+ ".getType()");
				System.err.println(je.getLocalizedMessage());
				je.printStackTrace();
				System.exit(1);
			} catch (CloneNotSupportedException cnse) {
				System.err.println("JexToFrames.convertExpres"
				+ "sion: CloneNotSupportedException for expre"
				+ "ssion.getType()");
				System.exit(1);
			}
			try {
				StringWriter sw= new StringWriter();
				expression.writeText(sw);
				frame.ssf("description",
				          new FILString(sw.toString()));
			} catch (IOException ioe) {
				System.err.println("JexToFrames:convertExpres"
				+ "sion: writeText threw an IOException!");
				System.err.println(ioe.getLocalizedMessage());
				ioe.printStackTrace();
				System.exit(1);
			}
			///////////////////////////////////////
			// = and :=: take a set of arguments //
			// to fill one slot                  //
			///////////////////////////////////////
			if (operator instanceof EQ
			 || operator instanceof INST_EQ)
			{
				UnorderedVector s= new UnorderedVector();
				try {
					s.addElement(convertSimpleExpression
					(expression.getSimpleExpression1()));
				} catch (UnsetValueException uve) {
					System.err.println("JexToFrames.conve"
					+ "rtExpression: UnsetValueException "
					+ "for expression.getSimpleExpression"
					+ "1()");
					System.exit(1);
				}
				try {
					s.addElement(convertSimpleExpression
					(expression.getSimpleExpression2()));
				} catch (UnsetValueException uve) {
					System.err.println("JexToFrames.conve"
					+ "rtExpression: UnsetValueException "
					+ "for expression.getSimpleExpression"
					+ "2()");
					System.exit(1);
				}
				frame.ssf(slot1Name,s);
				return frame;
			}
			/////////////////////////////////////
			// The others are binary operators //
			// and have two slots to fill      //
			/////////////////////////////////////
			Object slot1Filler= null;
			try {
				frame.ssf(slot1Name,
					slot1Filler= convertSimpleExpression
				        (expression.getSimpleExpression1()));
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convertExpres"
				+ "sion: UnsetValueException for expression.g"
				+ "etSimpleExpression1()");
				System.exit(1);
			}
			Object slot2Filler= null;
			try {
				frame.ssf(slot2Name,
					slot2Filler= convertSimpleExpression
				        (expression.getSimpleExpression2()));
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convertExpres"
				+ "sion: UnsetValueException for expression.g"
				+ "etSimpleExpression2()");
				System.exit(1);
			}
			/////////////////////////////////////
			// Set extra slots for comparators //
			/////////////////////////////////////
			if (operator instanceof GT || operator instanceof GE
			 || operator instanceof LT || operator instanceof LE)
			{
				//////////////////////////////////
				// A comparator is inclusive    //
				// iff the operator is >= or <= //
				//////////////////////////////////
				frame.ssf("inclusive",
				  new DefinedTypeInstance(DefinedType.
				      lookup("inclusive_or_exclusive"),
				  new EnumSymbol
				  (operator instanceof GE ||
				   operator instanceof LE
				   ? "inclusive" : "exclusive")));

				//////////////////////////////////////////
				// A comparator is discrete if both its //
				// greater and lesser slots are filled  //
				// with expressions of type integer     //
				//////////////////////////////////////////
				frame.ssf("discrete",
				  new DefinedTypeInstance(DefinedType.
				      lookup("discrete_or_continuous"),
				  new EnumSymbol(slot1Filler instanceof Frame
				              && isTypeDiscrete((Frame)
				                                slot1Filler)
				              && slot2Filler instanceof Frame
				              && isTypeDiscrete((Frame)
				                                slot2Filler)
				              ? "discrete" : "continuous")));

				////////////////////////////////////
				// A comparator is greater first  //
				// if it is > or >=, lesser first //
				// if it is < or <=               //
				////////////////////////////////////
				frame.ssf("which_way_round",
				  new DefinedTypeInstance
				     (DefinedType.lookup("way_round"),
				      new EnumSymbol(operator instanceof GT
				                  || operator instanceof GE
				      ? "greater_first" : "lesser_first")));
			}
			return frame;
		}
	}

	/********************************************************************/
	// static JexToFrames.convertSimpleExpression :
	//                    SimpleExpression -> Object
	/*********************************************************************
	** Converts a SimpleExpression to frames.
	** @param simpleExpression The simple expression to be converted.
	** @return The object into which the simpleExpression has been
	**         converted.
	*********************************************************************/

	private static Object convertSimpleExpression
	                     (SimpleExpression simpleExpression)
	{
		ListIterator li= simpleExpression.getSubSimpleExpressions();
		Term firstTerm= null;
		try {
			firstTerm= simpleExpression.getTerm();
		} catch (UnsetValueException uve) {
			System.err.println("JexToFrames.convertSimpleExpressi"
			+ "on: UnsetValueException for simpleExpression.getTe"
			+ "rm()");
			System.exit(1);
		}
		if (simpleExpression.countSubSimpleExpressions() == 0) {
			return convertTerm(firstTerm);
		} else {
			SubSimpleExpression firstSubSimpleExpression
			                    = (SubSimpleExpression)li.next();
			AddLikeOperator operator= null;
			try {
				operator= firstSubSimpleExpression.
				          getOperator();
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convertSimple"
				+ "Expression: UnsetValueException for firstS"
				+ "ubSimpleException.getOperator()");
				System.exit(1);
			}
			String fcName= null; String slotName= null;
			if (operator instanceof ADD
			 || operator instanceof SUB) {
				fcName= "addition";
				slotName= "addends";
			}
			if (operator instanceof OR) {
				fcName= "inclusive_disjunction";
				slotName= "disjuncts";
			}
			if (operator instanceof XOR) {
				fcName= "exclusive_disjunction";
				slotName= "disjuncts";
			}
			Frame frame= new Frame(FrameClass.lookup(fcName));
			try {
				frame.ssf("tipe",
				  convert(simpleExpression.getType()));
			} catch (IncompatibleTypeException ite) {
				System.err.println("JexToFrames.convertSimple"
				+ "Expression: Jex threw an IncompatibleTypeE"
				+ "xception for simpleExpression.getType()");
				System.err.println(ite.getLocalizedMessage());
				System.err.println("This is that stupid Jex "
				+ "1.9.3 error, so let's continue...");
				frame.ssf("tipe", new FILString
				                            (ite.toString()));
			} catch (ClassCastException cce) {
				System.err.println("JexToFrames.convertSimple"
				+ "Expression: Jex threw a ClassCastException"
				+ " for simpleExpression.getType()");
				System.err.println(cce.getLocalizedMessage());
				System.err.println("This is that stupid Jex "
				+ "1.9.5 error, so let's continue...");
				frame.ssf("tipe", new FILString
				                            (cce.toString()));
			} catch (JexException je) {
				System.err.println("JexToFrames.convertSimple"
				+ "Expression: " + je.getClass() + " for simp"
				+ "leExpression.getType()");
				System.err.println(je.getLocalizedMessage());
				je.printStackTrace();
				System.exit(1);
			} catch (CloneNotSupportedException cnse) {
				System.err.println("JexToFrames.convertSimple"
				+ "Expression: CloneNotSupportedException for"
				+ " simpleExpression.getType()");
				System.exit(1);
			}
			UnorderedVector terms= new UnorderedVector();
			terms.addElement(convertTerm(firstTerm));
			Object term= null;
			try {
				term= convertTerm(firstSubSimpleExpression.
			                          getTerm());
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convertSimple"
				+ "Expression: UnsetValueException for firstS"
				+ "ubSimpleException.getTerm()");
				System.exit(1);
			}
			if (operator instanceof SUB) {
				Frame minus= new Frame
				  (FrameClass.lookup("minus"));
				try {
					minus.setSlotFiller("operand",term);
				} catch (InvalidSlot is) {
					System.err.println("No slot \"operand"
					+ "\" of frame class \"minus\"");
					System.exit(1);
				}
				term= minus;
			}
			terms.addElement(term);
			while (li.hasNext()) {
				SubSimpleExpression subSimpleExpression
				= (SubSimpleExpression)li.next();
				try {
					term= convertTerm
					  (subSimpleExpression.getTerm());
				} catch (UnsetValueException uve) {
					System.err.println("JexToFrames.conve"
					+ "rtSimpleExpression: UnsetValueExce"
					+ "ption for subSimpleExpression.getT"
					+ "erm()");
					System.exit(1);
				}
				try {
					operator= subSimpleExpression.
					          getOperator();
				} catch (UnsetValueException uve) {
					System.err.println("JexToFrames.conve"
					+ "rtSimpleExpression: UnsetValueExce"
					+ "ption for subSimpleExpression.getO"
					+ "perator()");
					System.exit(1);
				}
				if (operator instanceof SUB) {
					Frame minus= new Frame
					  (FrameClass.lookup("minus"));
					try {
						minus.setSlotFiller("operand",
						                    term);
					} catch (InvalidSlot is) {
						System.err.println("No slot "
						+ "\"operand\" of frame class"
						+ "\"minus\"");
						System.exit(1);
					}
					term= minus;
				}
				terms.addElement(term);
			}
			try {
				frame.setSlotFiller(slotName,terms);
			} catch (InvalidSlot is) {
				System.err.println("No slot \"" + slotName
				+ "\" of frame class \"" + fcName + "\"!");
				System.exit(1);
			}
			return frame;
		}
	}

	/********************************************************************/
	// static JexToFrames.convertTerm : Term -> Object
	/*********************************************************************
	** Converts a Term to frames.
	** @param term The term to be converted.
	** @return The object into which the term has been converted.
	*********************************************************************/

	private static Object convertTerm (Term term) {
		ListIterator li= term.getSubTerms();
		Factor firstFactor= null;
		try {
			firstFactor= term.getFactor();
		} catch (UnsetValueException uve) {
			System.err.println("JexToFrames.convertTerm: UnsetVal"
			+ "ueException for term.getFactor()");
			System.exit(1);
		}
		if (! li.hasNext() /*term.countSubTerms() == 0*/) {
			return convertFactor(firstFactor);
		} else {
			SubTerm firstSubTerm= (SubTerm)li.next();
			MultiplicationLikeOperator operator= null;
			try {
				operator= firstSubTerm.getOperator();
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convertTerm: "
				+ "UnsetValueException for firstSubTerm.getOp"
				+ "erator()");
				System.exit(1);
			}
			String fcName= null; String slotName= null;
			if (operator instanceof MULT
			 || operator instanceof REAL_DIV) {
				fcName= "multiplication";
				slotName= "multiplicands";
			}
			if (operator instanceof AND) {
				fcName= "conjunction";
				slotName= "conjuncts";
			}
			if (operator instanceof COMPLEX_CONSTRUCTOR) {
				fcName= "complex_constructor";
				slotName= "entity_initialisers";
			}
			if (operator instanceof DIV) {
				fcName= "div_operation";
			}
			if (operator instanceof MOD) {
				fcName= "mod_operation";
			}
			Frame frame= new Frame(FrameClass.lookup(fcName));
			try {
				frame.setSlotFiller("tipe",
				  convert(term.getType()));
			} catch (IncompatibleTypeException ite) {
				System.err.println("JexToFrames.convertTerm: "
				+ "Jex threw an IncompatibleTypeException for"
				+ " term.getType()");
				System.err.println(ite.getLocalizedMessage());
				System.err.println("This is that stupid Jex "
				+ "1.9.3 error, so let's continue...");
				frame.ssf("tipe", new FILString
				                            (ite.toString()));
			} catch (ClassCastException cce) {
				System.err.println("JexToFrames.convertSimple"
				+ "Expression: Jex threw a ClassCastException"
				+ " for simpleExpression.getType()");
				System.err.println(cce.getLocalizedMessage());
				System.err.println("This is that stupid Jex "
				+ "1.9.5 error, so let's continue...");
				frame.ssf("tipe", new FILString
				                            (cce.toString()));
			} catch (InvalidSlot is) {
				System.err.println("No slot \"tipe\" of fra"
				+ "me class \"" + fcName + "\"!");
				System.exit(1);
			} catch (JexException je) {
				System.err.println("JexToFrames.convertTerm: "
				+ je.getClass() + " for term.getType()");
				System.err.println(je.getLocalizedMessage());
				je.printStackTrace();
				System.exit(1);
			} catch (CloneNotSupportedException cnse) {
				System.err.println("JexToFrames.convertTerm: "
				+ "CloneNotSupportedException for factor.getT"
				+ "ype()");
				System.exit(1);
			}
			if (operator instanceof MOD
			 || operator instanceof DIV) {
				try {
					frame.setSlotFiller("dividend",
					  convertFactor
					    (firstSubTerm.getFactor()));
				} catch (InvalidSlot is) {
					System.err.println("JexToFrames:conve"
					+ "rtTerm: No slot \"dividend\" of fr"
					+ "ame class " + fcName);
					System.exit(1);
				} catch (UnsetValueException uve) {
					System.err.println("JexToFrames.conve"
					+ "rtTerm: UnsetValueException for fi"
					+ "rstSubTerm.getFactor()");
					System.exit(1);
				}
				try {
					frame.ssf("divisor", convertFactor
					  (((SubTerm)li.next()).getFactor()));
				} catch (NoSuchElementException nsee) {
					System.err.println("Warning: " + frame
					+ "has no divisor!");
				} catch (UnsetValueException uve) {
					System.err.println("JexToFrames.conve"
					+ "rtTerm: UnsetValueException"
					+ "ption for getFactor()");
					System.exit(1);
				}
				if (li.hasNext()) {
					System.err.println("Can't handle DIV/"
					+ "MOD operation with more than two "
					+ "factors :-(");
					System.exit(1);
				}
				return frame;
			} //else:-
			UnorderedVector factors= new UnorderedVector();
			factors.addElement(convertFactor(firstFactor));
			Object factor= null;
			try {
				factor= convertFactor(firstSubTerm.
			                              getFactor());
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convertTerm: "
				+ "UnsetValueException for firstSubTerm.getFa"
				+ "ctor()");
				System.exit(1);
			}
			if (operator instanceof REAL_DIV) {
				Frame reciprocal= new Frame
				  (FrameClass.lookup("reciprocal"));
				try {
					reciprocal.setSlotFiller("operand",
					                         factor);
				} catch (InvalidSlot is) {
					System.err.println("No slot \"operand"
					+ "\" of frame class \"reciprocal\"");
					System.exit(1);
				}
				factor= reciprocal;
			}
			factors.addElement(factor);
			while (li.hasNext()) {
				SubTerm subTerm= (SubTerm)li.next();
				try {
					factor= convertFactor
					  (subTerm.getFactor());
				} catch (UnsetValueException uve) {
					System.err.println("JexToFrames.conve"
					+ "rtTerm: UnsetValueException for su"
					+ "bTerm.getFactor()");
					System.exit(1);
				}
				try {
					operator= subTerm.getOperator();
				} catch (UnsetValueException uve) {
					System.err.println("JexToFrames.conve"
					+ "rtTerm: UnsetValueException for su"
					+ "bTerm.getOperator()");
					System.exit(1);
				}
				if (operator instanceof REAL_DIV) {
					Frame reciprocal= new Frame
					  (FrameClass.lookup("reciprocal"));
					try {
						reciprocal.setSlotFiller
						           ("operand",factor);
					} catch (InvalidSlot is) {
						System.err.println("No slot "
						+ "\"operand\" of frame class"
						+ "\"reciprocal\"");
						System.exit(1);
					}
					factor= reciprocal;
				}
				factors.addElement(factor);
			}
			try {
				frame.setSlotFiller(slotName,factors);
			} catch (InvalidSlot is) {
				System.err.println("No slot \"" + slotName
				+ "\" of frame class \"" + fcName + "\"!");
				System.exit(1);
			}
			return frame;
		}
	}

	/********************************************************************/
	// static JexToFrames.convertFactor : Factor -> Object
	/*********************************************************************
	** Converts a Factor to frames.
	** @param factor The factor to be converted.
	** @return The object into which the factor has been converted.
	*********************************************************************/

	private static Object convertFactor (Factor factor)
	{
		if (! factor.hasSimpleFactor2()) {
			try {
				return convert(factor.getSimpleFactor1());
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convertFactor"
				+ ": UnsetValueException for factor.getSimple"
				+ "Factor1()");
				System.exit(1); return null;
			}
		} else {
			Frame frame= new Frame(FrameClass.lookup
			                       ("exponentiation"));
			try {
				frame.setSlotFiller("tipe",
				  convert(factor.getType()));
			} catch (InvalidSlot is) {
				System.err.println("No slot \"tipe\" of fra"
				+ "me class \"exponentiation\"!");
				System.exit(1);
			} catch (JexException je) {
				System.err.println("JexToFrames.convertFactor"
				+ ": " + je.getClass() + " for factor.getType"
				+ "()");
				System.err.println(je.getLocalizedMessage());
				je.printStackTrace();
				System.exit(1);
			} catch (CloneNotSupportedException cnse) {
				System.err.println("JexToFrames.convertFactor"
				+ ": CloneNotSupportedException for factor.ge"
				+ "tType()");
				System.exit(1);
			}
			try {
				frame.setSlotFiller("base",convert
				  (factor.getSimpleFactor1()));
			} catch (InvalidSlot is) {
				System.err.println("No slot \"base\" of frame"
				+ " class \"exponentiation\"!");
				System.exit(1);
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convertFactor"
				+ ": UnsetValueException for factor.getSimple"
				+ "Factor1()");
				System.exit(1);
			}
			try {
				frame.setSlotFiller("exponent",convert
				  (factor.getSimpleFactor2()));
			} catch (InvalidSlot is) {
				System.err.println("No slot \"exponent\" of "
				+ "frame class \"exponentiation\"!");
				System.exit(1);
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convertFactor"
				+ ": UnsetValueException for factor.getSimple"
				+ "Factor2()");
				System.exit(1);
			}
			return frame;
		}
	}

	/********************************************************************/
	// static JexToFrames.convertQualifiedFactor :
	//                    QualifiedFactor -> Object
	/*********************************************************************
	** Converts a QualifiedFactor to frames.
	** @param qualifiedFactor The qualified factor to be converted.
	** @return The object into which the qualifiedFactor has been
	**         converted.
	*********************************************************************/

	private static Object convertQualifiedFactor
	                      (QualifiedFactor qualifiedFactor)
	{
		Object qualifiableFactor= null;
		try {
			qualifiableFactor= convert(qualifiedFactor.
			                           getQualifiableFactor());
		} catch (UnsetValueException uve) {
			System.err.println("JexToFrames.convertQualifiedFacto"
			+ "r: UnsetValueException for qualifiedFactor.getQual"
			+ "ifiableFactor()");
			System.exit(1);
		}

		if (! qualifiedFactor.hasQualifier()) {
			return qualifiableFactor;
		}
		else {
			try {
				return convertQualifier(qualifiableFactor,
			               qualifiedFactor.getQualifier());
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convertQualif"
				+ "iedFactor: UnsetValueException thrown for "
				+ "qualifiedFactor.getQualifier()");
				System.exit(1); return null;
			}
		}
	}

	/********************************************************************/
	// static JexToFrames.convertQualifier :
	//                    Object x Qualifier -> Frame
	/*********************************************************************
	** Converts a qualified factor and a Jex Qualifier object to a
	** qualifier frame.
	** @param qualifiedFactor The qualified factor (FIL object).
	** @param qualifier The Qualifier object (Jex object).
	** @return A qualifier frame.
	*********************************************************************/

	private static Frame convertQualifier (Object qualifiedFactor,
	                                       Qualifier qualifier)
	{
		Frame qualifierFrame= null;

		////////////////////////////////
		// Attribute qualifiers - x.a //
		////////////////////////////////
		if (qualifier instanceof AttributeQualifier) {
			qualifierFrame= new Frame("attribute_qualifier");
			qualifierFrame.ssf("qualified_factor",
			                    qualifiedFactor);
			try {
				qualifierFrame.ssf("attribute_reference",
			                    convertRef(((AttributeQualifier)
			                      qualifier).getAttributeRef()));
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convertQualif"
				+ "ier: UnsetValueException thrown for ((Attr"
				+ "ibuteQualifier)qualifier.getAttributeRef()"
				); System.exit(1);
			}
		}
		////////////////////////////
		// Group qualifiers - x\e //
		////////////////////////////
		if (qualifier instanceof GroupQualifier) {
			qualifierFrame= new Frame("group_qualifier");
			qualifierFrame.ssf("qualified_factor",
			                    qualifiedFactor);
			try {
				qualifierFrame.ssf("entity_reference",
			                    convertRef(((GroupQualifier)
			                      qualifier).getEntityRef()));
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convertQualif"
				+ "ier: UnsetValueException thrown for ((Grou"
				+ "pQualifier)qualifier.getEntityRef()");
				System.exit(1);
			}
		}
		/////////////////////////////////////
		// Index qualifiers - [i] or [i:j] //
		/////////////////////////////////////
		if (qualifier instanceof IndexQualifier) {
			qualifierFrame= new Frame("index_qualifier");
			try {
				qualifierFrame.ssf("index1",
			                    convert(((IndexQualifier)qualifier
			                      ).getNumericExpression1()));
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convertQualif"
				+ "ier: UnsetValueException thrown for ((Inde"
				+ "xQualifier)qualifier.getNumericExpression1"
				+ "()");
				System.exit(1);
			}
			try {
				qualifierFrame.ssf("index2",
			                    convert(((IndexQualifier)qualifier
			                      ).getNumericExpression2()));
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convertQualif"
				+ "ier: UnsetValueException thrown for ((Inde"
				+ "xQualifier)qualifier.getNumericExpression2"
				+ "()");
				System.exit(1);
			}
		}
		//////////////////////////////////////
		// Set general qualifier attributes //
		//////////////////////////////////////
		qualifierFrame.ssf("qualified_factor", qualifiedFactor);
		try {
			qualifierFrame.ssf("tipe",
			                   convert(qualifier.getType()));
		} catch (IncompatibleTypeException ite) {
			System.err.println("JexToFrames.convertQualifier: Jex"
			+ " threw an IncompatibleTypeException for qualifier."
			+ "getType()");
			System.err.println(ite.getLocalizedMessage());
			System.err.println("This is that stupid Jex 1.9.3 err"
			+ "or, so let's continue...");
			qualifierFrame.ssf("tipe",
			                   new FILString(ite.toString()));
		} catch (ClassCastException cce) {
			System.err.println("JexToFrames.convertQualifier: Jex"
			+ " threw a ClassCastException for qualifier.getType("
			+ ")");
			System.err.println(cce.getLocalizedMessage());
			System.err.println("This is that stupid Jex 1.9.3 err"
			+ "or, so let's continue...");
			qualifierFrame.ssf("tipe",
			                   new FILString(cce.toString()));
		} catch (TurnedOffMethodException tome) {
			System.err.println("Warning: JexToFrames.convertQuali"
			+ "fier: TurnedOffMethodException for (" + qualifier.
			getClass() + ")qualifier.getType()");
		} catch (JexException je) {
			System.err.println("JexToFrames.convertQualifier: "
			+ je.getClass() + " thrown for qualifier.getType()");
			System.err.println(je.getLocalizedMessage());
			je.printStackTrace();
			System.exit(1);
		} catch (CloneNotSupportedException cnse) {
			System.err.println("JexToFrames.convertQualifier: Clo"
			+ "neNotSupportedException for qualifier.getType()");
			System.exit(1);
		}
		////////////////////////////////////////////////
		// If there's another qualifier in the chain, //
		// it's an outer wrapper to this one.         //
		////////////////////////////////////////////////
		if (qualifier.hasQualifier()) {
			try {
				return convertQualifier(qualifierFrame,
			               qualifier.getQualifier());
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convertQualif"
				+ "ier: UnsetValueException thrown for qualif"
				+ "ier.getQualifier()");
				System.exit(1);
			}
		}
		return qualifierFrame;
	}

	/********************************************************************/
	// static JexToFrames.convertUserDefinedFunctionCall :
	//                    Object x ListIterator -> Frame
	/*********************************************************************
	** Converts a user-defined function reference and a ListIterator of
	** actual parameters to a function_call frame.
	** qualifier frame.
	** @param function The user-defined function (FIL object).
	** @param actualParametersListIterator A ListIterator of actual
	**                                     parameters (Jex objects).
	** @return A function_call frame.
	*********************************************************************/

	private static Frame convertUserDefinedFunctionCall (Object function,
	                     ListIterator actualParametersListIterator)
	{
		Frame functionCallFrame= new Frame("function_call");
		functionCallFrame.ssf("function_called", function);

		OrderedVector actualParameters= new OrderedVector();
		while (actualParametersListIterator.hasNext()) {
			actualParameters.addElement(JexToFrames.convert
			               (actualParametersListIterator.next()));
		}
		functionCallFrame.ssf("actual_parameters", actualParameters);

		functionCallFrame.ssf("is_variable", Logical.TRUE);
		try {
			functionCallFrame.ssf("tipe", ((Frame)function).
		    	                        getSlotFiller("return_type"));
		} catch (InvalidSlot is) {
			functionCallFrame.ssf("tipe", "entity_type#197.gsf: "
			+ "InvalidSlot \"return_type\"");
		}
		return functionCallFrame;
	}

	/********************************************************************/
	// static JexToFrames.convertInterval : Interval -> Object
	/*********************************************************************
	** Converts a Jex Interval object to an "interval" frame.
	** @param interval The interval to be converted.
	** @return The frame into which the interval has been converted.
	*********************************************************************/

	private static Frame convertInterval (Interval interval) {
		Frame frame= new Frame(FrameClass.lookup("interval"));

		frame.ssf("tipe", convert(interval.getType()));
		try {
			frame.ssf("lower_bound",
			          convert(interval.getLowerExpression()));
		} catch (UnsetValueException uve) {
			System.err.println("JexToFrames.convertInterval: "
			+ "UnsetValueException for interval."
			+ "getLowerExpression()");
			System.exit(1);
		}
		try {
			frame.ssf("lower_inclusive", new DefinedTypeInstance(
			         DefinedType.lookup("inclusive_or_exclusive"),
			  new EnumSymbol(interval.getLowerToMiddleOperator()
			      instanceof LE ? "inclusive" : "exclusive")));
		} catch (UnsetValueException uve) {
			System.err.println("JexToFrames.convertInterval: "
			+ "UnsetValueException for interval."
			+ "getLowerToMiddleOperator()");
			System.exit(1);
		}
		Object variable= null;
		try {
			frame.ssf("variable",
			  variable= convert(interval.getMiddleExpression()));
		} catch (UnsetValueException uve) {
			System.err.println("JexToFrames.convertInterval: "
			+ "UnsetValueException for interval."
			+ "getMiddleExpression()");
			System.exit(1);
		}
		/////////////////////////////////////////////
		// An interval is discrete if its variable //
		// is an expression of type integer        //
		/////////////////////////////////////////////
		frame.ssf("discrete", new DefinedTypeInstance(DefinedType.
		                      lookup("discrete_or_continuous"), new
		                      EnumSymbol(variable instanceof Frame
		                            && isTypeDiscrete((Frame)variable)
		                      ? "discrete" : "continuous")));
		try {
			frame.ssf("upper_inclusive", new DefinedTypeInstance(
			         DefinedType.lookup("inclusive_or_exclusive"),
			  new EnumSymbol(interval.getMiddleToUpperOperator()
			      instanceof LE ? "inclusive" : "exclusive")));
		} catch (UnsetValueException uve) {
			System.err.println("JexToFrames.convertInterval: "
			+ "UnsetValueException for interval."
			+ "getMiddleToUpperOperator()");
			System.exit(1);
		}
		try {
			frame.ssf("upper_bound",
			  convert(interval.getUpperExpression()));
		} catch (UnsetValueException uve) {
			System.err.println("JexToFrames.convertInterval: "
			+ "UnsetValueException for interval."
			+ "getUpperExpression()");
			System.exit(1);
		}
		return frame;
	}

	/********************************************************************/
	// static JexToFrames.convertUnaryOperationFactor :
	//                    UnaryOperationFactor -> Object
	/*********************************************************************
	** Converts a Jex UnaryOperationFactor object to a frame.
	** @param uof The unary operation factor to be converted.
	** @return The frame into which the unary operation factor has been
	**         converted.
	*********************************************************************/

	private static Object convertUnaryOperationFactor
	                      (UnaryOperationFactor uof)
	{
		UnaryOperator operator= null; Object operand= null;
		try {
			operator= uof.getUnaryOperator();
			operand= convert(uof.getSimpleSubFactor());
		} catch (UnsetValueException uve) {
			System.err.println("JexToFrames.convertUnaryOperation"
			+ "Factor(" + uof + "): UnsetValueException for "
			+ "uof.getUnaryOperator()");
			System.exit(1); return null;
		}

		if (operator instanceof PLUS)
			return operand;

		Frame frame= new Frame(FrameClass.lookup
		             (operator instanceof MINUS ? "minus"
		            /*operator instanceof NOT*/ : "negation"));
		frame.ssf("operand", operand);
		try {
			frame.ssf("tipe", convert(uof.getType()));
		} catch (IncompatibleTypeException ite) {
			System.err.println("JexToFrames.convertUnaryOperation"
			+ "Factor: Jex threw an IncompatibleTypeException for"
			+ " uof.getType()");
			System.err.println(ite.getLocalizedMessage());
			System.err.println("This is that stupid Jex "
			+ "1.9.3 error, so let's continue...");
			frame.ssf("tipe", new FILString(ite.toString()));
		} catch (ClassCastException cce) {
			System.err.println("JexToFrames.convertSimple"
			+ "Expression: Jex threw a ClassCastException"
			+ " for simpleExpression.getType()");
			System.err.println(cce.getLocalizedMessage());
			System.err.println("This is that stupid Jex "
			+ "1.9.5 error, so let's continue...");
			frame.ssf("tipe", new FILString(cce.toString()));
		} catch (JexException je) {
			System.err.println("JexToFrames.convertUnaryOperation"
			+ "Factor: " + je.getClass() + " for uof.getType()");
			System.err.println(je.getLocalizedMessage());
			je.printStackTrace();
			System.exit(1);
		} catch (CloneNotSupportedException cnse) {
			System.err.println("JexToFrames.convertUnaryOperation"
			+ "Factor: CloneNotSupportedException for uof."
			+ "getType()");
			System.exit(1);
		}
		return frame;
	}

	/********************************************************************/
	// static JexToFrames.convertSetType : SetType -> Object
	/*********************************************************************
	** Converts a Jex SetType object to a "set_type" frame.
	** @param setType The set-type to be converted.
	** @return The frame into which the set-type has been converted.
	*********************************************************************/

	private static Frame convertSetType (SetType setType) {
		Frame frame= new Frame(FrameClass.lookup("set_type"));

		try {
			frame.ssf("element_type",
			          convert(setType.getBaseType()));
		} catch (UnsetValueException uve) {
			System.err.println("JexToFrames.convert"
			+ "SetType: UnsetValueException for "
			+ "setType.getBaseType()");
			System.exit(1);
		}
		BoundSpec boundSpec= null;
		try {
			boundSpec= setType.getBoundSpec();
		} catch (UnsetValueException uve) {
			System.err.println("JexToFrames.convert"
			+ "SetType: UnsetValueException for "
			+ "setType.getBoundSpec()");
			System.exit(1);
		}
		if (boundSpec != null) {
			try {
				frame.ssf("lower_bound",
				          convert(boundSpec.getLowerBound()));
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convert"
				+ "SetType: UnsetValueException for "
				+ "boundSpec.getLowerBound()");
				System.exit(1);
			}
			try {
				frame.ssf("upper_bound",
				          convert(boundSpec.getUpperBound()));
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convert"
				+ "SetType: UnsetValueException for "
				+ "boundSpec.getUpperBound()");
				System.exit(1);
			}
		} else {
			Frame zero
			= new Frame(FrameClass.lookup("integer_literal"));
			zero.ssf("tipe", new DefinedTypeInstance
			                   (DefinedType.lookup("simple_type"),
			                    new EnumSymbol("integer_type")));
			zero.ssf("is_variable", Logical.FALSE);
			zero.ssf("literal_value", new Integer(0));
			frame.ssf("lower_bound", zero);
			Frame indet
			= new Frame(FrameClass.lookup("built_in_constant"));
			indet.ssf("built_in_value", new DefinedTypeInstance
			                   (DefinedType.lookup("built_in"),
			                    new EnumSymbol("indeterminate")));
			indet.ssf("tipe", new DefinedTypeInstance
			            (DefinedType.lookup("indeterminate_type"),
			             new EnumSymbol("indeterminate")));
			indet.ssf("is_variable", Logical.FALSE);
			frame.ssf("upper_bound", indet);

		}
		return frame;
	}

	/********************************************************************/
	// static JexToFrames.convertGeneralSetType : GeneralSetType -> Object
	/*********************************************************************
	** Converts a Jex GeneralSetType object to a "set_type" frame.
	** @param setType The set-type to be converted.
	** @return The frame into which the set-type has been converted.
	*********************************************************************/

	private static Frame convertGeneralSetType (GeneralSetType setType) {
		Frame frame= new Frame(FrameClass.lookup("set_type"));

		try {
			frame.ssf("element_type",
			          convert(setType.getParameterType()));
		} catch (UnsetValueException uve) {
			System.err.println("JexToFrames.convert"
			+ "GeneralSetType: UnsetValueException for "
			+ "setType.getBaseType()");
			System.exit(1);
		}
		BoundSpec boundSpec= null;
		try {
			boundSpec= setType.getBoundSpec();
		} catch (UnsetValueException uve) {
			System.err.println("JexToFrames.convert"
			+ "GeneralSetType: UnsetValueException for "
			+ "setType.getBoundSpec()");
			System.exit(1);
		}
		if (boundSpec != null) {
			try {
				frame.ssf("lower_bound",
				          convert(boundSpec.getLowerBound()));
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convert"
				+ "GeneralSetType: UnsetValueException for "
				+ "boundSpec.getLowerBound()");
				System.exit(1);
			}
			try {
				frame.ssf("upper_bound",
				          convert(boundSpec.getUpperBound()));
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convert"
				+ "GeneralSetType: UnsetValueException for "
				+ "boundSpec.getUpperBound()");
				System.exit(1);
			}
		} else {
			Frame zero
			= new Frame(FrameClass.lookup("integer_literal"));
			zero.ssf("tipe", new DefinedTypeInstance
			                   (DefinedType.lookup("simple_type"),
			                    new EnumSymbol("integer_type")));
			zero.ssf("is_variable", Logical.FALSE);
			zero.ssf("literal_value", new Integer(0));
			frame.ssf("lower_bound", zero);
			Frame indet
			= new Frame(FrameClass.lookup("built_in_constant"));
			indet.ssf("built_in_value", new DefinedTypeInstance
			                   (DefinedType.lookup("built_in"),
			                    new EnumSymbol("indeterminate")));
			indet.ssf("tipe", new DefinedTypeInstance
			            (DefinedType.lookup("indeterminate_type"),
			             new EnumSymbol("indeterminate")));
			indet.ssf("is_variable", Logical.FALSE);
			frame.ssf("upper_bound", indet);

		}
		return frame;
	}

	/********************************************************************/
	// static JexToFrames.convertBagType : BagType -> Object
	/*********************************************************************
	** Converts a Jex BagType object to a "bag_type" frame.
	** @param bagType The bag-type to be converted.
	** @return The frame into which the bag-type has been converted.
	*********************************************************************/

	private static Frame convertBagType (BagType bagType) {
		Frame frame= new Frame(FrameClass.lookup("bag_type"));

		try {
			frame.ssf("element_type",
			          convert(bagType.getBaseType()));
		} catch (UnsetValueException uve) {
			System.err.println("JexToFrames.convert"
			+ "BagType: UnsetValueException for "
			+ "bagType.getBaseType()");
			System.exit(1);
		}
		BoundSpec boundSpec= null;
		try {
			boundSpec= bagType.getBoundSpec();
		} catch (UnsetValueException uve) {
			System.err.println("JexToFrames.convert"
			+ "BagType: UnsetValueException for "
			+ "bagType.getBoundSpec()");
			System.exit(1);
		}
		if (boundSpec != null) {
			try {
				frame.ssf("lower_bound",
				          convert(boundSpec.getLowerBound()));
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convert"
				+ "BagType: UnsetValueException for "
				+ "boundSpec.getLowerBound()");
				System.exit(1);
			}
			try {
				frame.ssf("upper_bound",
				          convert(boundSpec.getUpperBound()));
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convert"
				+ "BagType: UnsetValueException for "
				+ "boundSpec.getUpperBound()");
				System.exit(1);
			}
		} else {
			Frame zero
			= new Frame(FrameClass.lookup("integer_literal"));
			zero.ssf("tipe", new DefinedTypeInstance
			                   (DefinedType.lookup("simple_type"),
			                    new EnumSymbol("integer_type")));
			zero.ssf("is_variable", Logical.FALSE);
			zero.ssf("literal_value", new Integer(0));
			frame.ssf("lower_bound", zero);
			Frame indet
			= new Frame(FrameClass.lookup("built_in_constant"));
			indet.ssf("built_in_value", new DefinedTypeInstance
			                   (DefinedType.lookup("built_in"),
			                    new EnumSymbol("indeterminate")));
			indet.ssf("tipe", new DefinedTypeInstance
			            (DefinedType.lookup("indeterminate_type"),
			             new EnumSymbol("indeterminate")));
			indet.ssf("is_variable", Logical.FALSE);
			frame.ssf("upper_bound", indet);

		}
		return frame;
	}

	/********************************************************************/
	// static JexToFrames.convertGeneralBagType : GeneralBagType -> Object
	/*********************************************************************
	** Converts a Jex GeneralBagType object to a "bag_type" frame.
	** @param bagType The bag-type to be converted.
	** @return The frame into which the bag-type has been converted.
	*********************************************************************/

	private static Frame convertGeneralBagType (GeneralBagType bagType) {
		Frame frame= new Frame(FrameClass.lookup("bag_type"));

		try {
			frame.ssf("element_type",
			          convert(bagType.getParameterType()));
		} catch (UnsetValueException uve) {
			System.err.println("JexToFrames.convert"
			+ "GeneralBagType: UnsetValueException for "
			+ "bagType.getBaseType()");
			System.exit(1);
		}
		BoundSpec boundSpec= null;
		try {
			boundSpec= bagType.getBoundSpec();
		} catch (UnsetValueException uve) {
			System.err.println("JexToFrames.convert"
			+ "GeneralBagType: UnsetValueException for "
			+ "bagType.getBoundSpec()");
			System.exit(1);
		}
		if (boundSpec != null) {
			try {
				frame.ssf("lower_bound",
				          convert(boundSpec.getLowerBound()));
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convert"
				+ "GeneralBagType: UnsetValueException for "
				+ "boundSpec.getLowerBound()");
				System.exit(1);
			}
			try {
				frame.ssf("upper_bound",
				          convert(boundSpec.getUpperBound()));
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convert"
				+ "GeneralBagType: UnsetValueException for "
				+ "boundSpec.getUpperBound()");
				System.exit(1);
			}
		} else {
			Frame zero
			= new Frame(FrameClass.lookup("integer_literal"));
			zero.ssf("tipe", new DefinedTypeInstance
			                   (DefinedType.lookup("simple_type"),
			                    new EnumSymbol("integer_type")));
			zero.ssf("is_variable", Logical.FALSE);
			zero.ssf("literal_value", new Integer(0));
			frame.ssf("lower_bound", zero);
			Frame indet
			= new Frame(FrameClass.lookup("built_in_constant"));
			indet.ssf("built_in_value", new DefinedTypeInstance
			                   (DefinedType.lookup("built_in"),
			                    new EnumSymbol("indeterminate")));
			indet.ssf("tipe", new DefinedTypeInstance
			            (DefinedType.lookup("indeterminate_type"),
			             new EnumSymbol("indeterminate")));
			indet.ssf("is_variable", Logical.FALSE);
			frame.ssf("upper_bound", indet);

		}
		return frame;
	}

	/********************************************************************/
	// static JexToFrames.convertListType : ListType -> Object
	/*********************************************************************
	** Converts a Jex ListType object to a "list_type" frame.
	** @param listType The list-type to be converted.
	** @return The frame into which the list-type has been converted.
	*********************************************************************/

	private static Frame convertListType (ListType listType) {
		Frame frame= new Frame(FrameClass.lookup("list_type"));

		try {
			frame.ssf("element_type",
			          convert(listType.getBaseType()));
		} catch (UnsetValueException uve) {
			System.err.println("JexToFrames.convert"
			+ "ListType: UnsetValueException for "
			+ "listType.getBaseType()");
			System.exit(1);
		}
		BoundSpec boundSpec= null;
		try {
			boundSpec= listType.getBoundSpec();
		} catch (UnsetValueException uve) {
			System.err.println("JexToFrames.convert"
			+ "ListType: UnsetValueException for "
			+ "listType.getBoundSpec()");
			System.exit(1);
		}
		if (boundSpec != null) {
			try {
				frame.ssf("lower_bound",
				          convert(boundSpec.getLowerBound()));
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convert"
				+ "ListType: UnsetValueException for "
				+ "boundSpec.getLowerBound()");
				System.exit(1);
			}
			try {
				frame.ssf("upper_bound",
				          convert(boundSpec.getUpperBound()));
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convert"
				+ "ListType: UnsetValueException for "
				+ "boundSpec.getUpperBound()");
				System.exit(1);
			}
		} else {
			Frame zero
			= new Frame(FrameClass.lookup("integer_literal"));
			zero.ssf("tipe", new DefinedTypeInstance
			                   (DefinedType.lookup("simple_type"),
			                    new EnumSymbol("integer_type")));
			zero.ssf("is_variable", Logical.FALSE);
			zero.ssf("literal_value", new Integer(0));
			frame.ssf("lower_bound", zero);
			Frame indet
			= new Frame(FrameClass.lookup("built_in_constant"));
			indet.ssf("built_in_value", new DefinedTypeInstance
			                   (DefinedType.lookup("built_in"),
			                    new EnumSymbol("indeterminate")));
			indet.ssf("tipe", new DefinedTypeInstance
			            (DefinedType.lookup("indeterminate_type"),
			             new EnumSymbol("indeterminate")));
			indet.ssf("is_variable", Logical.FALSE);
			frame.ssf("upper_bound", indet);

		}
		frame.ssf("is_unique", listType.isUnique() ? Logical.TRUE
		                                           : Logical.FALSE);
		return frame;
	}

	/********************************************************************/
	// static JexToFrames.convertGeneralListType : GeneralListType
	//                                             -> Object
	/*********************************************************************
	** Converts a Jex GeneralListType object to a "list_type" frame.
	** @param listType The list-type to be converted.
	** @return The frame into which the list-type has been converted.
	*********************************************************************/

	private static Frame convertGeneralListType (GeneralListType listType)
	{
		Frame frame= new Frame(FrameClass.lookup("list_type"));

		try {
			frame.ssf("element_type",
			          convert(listType.getParameterType()));
		} catch (UnsetValueException uve) {
			System.err.println("JexToFrames.convert"
			+ "GeneralListType: UnsetValueException for "
			+ "listType.getBaseType()");
			System.exit(1);
		}
		BoundSpec boundSpec= null;
		try {
			boundSpec= listType.getBoundSpec();
		} catch (UnsetValueException uve) {
			System.err.println("JexToFrames.convert"
			+ "GeneralListType: UnsetValueException for "
			+ "listType.getBoundSpec()");
			System.exit(1);
		}
		if (boundSpec != null) {
			try {
				frame.ssf("lower_bound",
				          convert(boundSpec.getLowerBound()));
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convert"
				+ "GeneralListType: UnsetValueException for "
				+ "boundSpec.getLowerBound()");
				System.exit(1);
			}
			try {
				frame.ssf("upper_bound",
				          convert(boundSpec.getUpperBound()));
			} catch (UnsetValueException uve) {
				System.err.println("JexToFrames.convert"
				+ "GeneralListType: UnsetValueException for "
				+ "boundSpec.getUpperBound()");
				System.exit(1);
			}
		} else {
			Frame zero
			= new Frame(FrameClass.lookup("integer_literal"));
			zero.ssf("tipe", new DefinedTypeInstance
			                   (DefinedType.lookup("simple_type"),
			                    new EnumSymbol("integer_type")));
			zero.ssf("is_variable", Logical.FALSE);
			zero.ssf("literal_value", new Integer(0));
			frame.ssf("lower_bound", zero);
			Frame indet
			= new Frame(FrameClass.lookup("built_in_constant"));
			indet.ssf("built_in_value", new DefinedTypeInstance
			                   (DefinedType.lookup("built_in"),
			                    new EnumSymbol("indeterminate")));
			indet.ssf("tipe", new DefinedTypeInstance
			            (DefinedType.lookup("indeterminate_type"),
			             new EnumSymbol("indeterminate")));
			indet.ssf("is_variable", Logical.FALSE);
			frame.ssf("upper_bound", indet);

		}
		frame.ssf("is_unique", listType.isUnique() ? Logical.TRUE
		                                           : Logical.FALSE);
		return frame;
	}

	/********************************************************************/
	// static JexToFrames.isTypeDiscrete : Frame -> boolean
	/*********************************************************************
	** Returns true if the given frame has a discrete (i.e. integer)
	** type, or false if it has a continuous (i.e. real-number) type.
	** If it cannot determine whether the object is of a discrete type or 
	** a continuous type, it prints a warning and returns false.
	** @return Whether this frame has a discrete type.
	*********************************************************************/

	private static boolean isTypeDiscrete (Frame frame) {
		Object type= null;
		try {
			type= frame.getSlotFiller("tipe");
		} catch (InvalidSlot is) {
			System.err.println("JexToFrames.isTypeDiscrete: frame"
			+ ".getSlotFiller(\"tipe\") threw an InvalidSlot ("
			+ "frame == " + frame + ")");
			System.exit(1);
		}
		if (type instanceof DefinedTypeInstance)
			return type.toString().equals("integer_type");
		else if (type instanceof Frame) {
			Object tut= null;
			try {
				tut= ((Frame)type).getSlotFiller("tipe");
			} catch (InvalidSlot is) {
				System.err.println("Warning: JexToFrames.isTy"
				+ "peDiscrete: type.getSlotFiller(\"tipe\") "
				+ "threw an InvalidSlot (type == " + type
				+ ") - this method should be rewritten");
			}
			if (tut != null) {
				return tut.toString().equals("integer_type");
			}
		}
		System.err.println("JexToFrames.isTypeDiscrete failed for "
		+ frame + " - assuming continuous type.");
		return false;
	}

	/********************************************************************/
	// static JexToFrames.writeClass : ObjectOutput -> void
	/*********************************************************************
	** Serialises the static data members (class variables) of
	** JexToFrames.
	** @param oo An output stream to save the class data to.
	** @exception IOException If an I/O exception occurs.
	*********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_cache);
	}

	/********************************************************************/
	// static JexToFrames.readClass : ObjectInputStream -> void
	/*********************************************************************
	** Deserialises the static data members (class variables) of
	** JexToFrames.
	** @param oi An input stream to save the class data to.
	** @exception IOException If an I/O exception occurs.
	*********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_cache= (Hashtable)oi.readObject();
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the st"
			+ "upid compiler is just too dumb to realise this!");
		}
	}
}
