      ////////////////////////////////////////
     // THE CONSTRAINT-UNDERSTANDING SYSTEM //
    //  (Version 2)                        ///
   // Written by Andrew Broad             ////
  //                                     /////
 // PrettyPrint.java (2001:12:06 17:14) //////
/////////////////////////////////////////////
// Methods of PrettyPrint:             /////
// - PrettyPrint constructor           ////
// - generateDescription               ///
// - prettyPrint                       //
////////////////////////////////////////

import java.util.*;
import java.io.*;

import jex.*;

/*****************************************************************************
** The class `PrettyPrint' encapsulates the algorithm for generating textual
** descriptions of frames (particularly HLCs).
*****************************************************************************/

public class PrettyPrint {
	private CaseLibrary _prettyPrintCaseLibrary;

	/********************************************************************/
	// PrettyPrint constructor : CaseLibrary ->
	/*********************************************************************
	** Constructs an instance of PrettyPrint with a given pretty-print
	** case library.
	** @param prettyPrintCaseLibrary The pretty-print case library.
	*********************************************************************/

	public PrettyPrint (CaseLibrary prettyPrintCaseLibrary) {
		_prettyPrintCaseLibrary= prettyPrintCaseLibrary;
	}

	/********************************************************************/
	// PrettyPrint.generateDescription : (Frame x boolean) -> String
	/*********************************************************************
	** Generates a natural language description for a given HLC. Sets the
	** description slot of the HLC frame if the second parameter is true.
	** @param HLC The frame representing the HLC.
	** @param tag Whether to set the frame's description slot.
	** @return A natural language description of the HLC.
	*********************************************************************/

	public String generateDescription (Frame HLC, boolean tag) {
		String text= "";
		try {
			Object description= HLC.getSlotFiller("description");
			if (description != null)
				return description.toString();
		} catch (InvalidSlot is) {}

		Vector ppCases= _prettyPrintCaseLibrary.retrieve(HLC);
		if (ppCases.size() >= 1) {
			Case ppCase= CaseLibrary.selectMostSpecific(ppCases);

			if (ppCase.match(HLC)) {
				Frame pp=(Frame)ppCase.extract().elementAt(0);
				Vector textV= (Vector)pp.gsf("text");
				Enumeration textE= textV.elements();
				while (textE.hasMoreElements()) {
					text+= prettyPrint
					       (textE.nextElement());
				}
				if (tag) HLC.ssf("description",
				                 new FILString(text));
				Frame.remove(pp);
			}
		}
		if (text.equals("") || text.equals("null")) {
			text= HLC.toString();
		}
		return text;
	}

	/********************************************************************/
	// ExtractionEngine.prettyPrint : Object -> String
	/*********************************************************************
	** Pretty-prints the given object. If it's a frame, it calls
	** generateDescription. If it's an aggregation, it recurses for each
	** element. Otherwise, it uses toString.
	** @param e The element to be pretty-printed.
	** @return A pretty-printing of the element.
	*********************************************************************/

	private String prettyPrint (Object e) {
		String text;

		if (e instanceof Frame) {
			text= generateDescription((Frame)e, false);
			if (FrameClass.lookup("higher_level_constraint").
			    subsumes(((Frame)e).getFrameClass())
			 && ! text.equals(e.toString()))
			{
				text= e.toString(); //= e + " (" + text + ")";
			}
			return text;
		}
		if (e instanceof UnorderedVector) {
			text= "{";
			Enumeration elements= ((UnorderedVector)e).elements();
			while (elements.hasMoreElements()) {
				text+= prettyPrint(elements.nextElement());
				if (elements.hasMoreElements())
					text+= ", ";
			}
			return text + "}";
		}
		if (e instanceof OrderedVector) {
			text= "[";
			Enumeration elements= ((OrderedVector)e).elements();
			while (elements.hasMoreElements()) {
				text+= prettyPrint(elements.nextElement());
				if (elements.hasMoreElements())
					text+= ", ";
			}
			return text + "]";
		}
		if (e == null)
			return "null";
		else
			return e.toString();
	}
}
