      ////////////////////////////////////////
     // THE CONSTRAINT-UNDERSTANDING SYSTEM //
    //  (Version 2)                        ///
   // Written by Andrew Broad             ////
  //                                     /////
 // Slot.java (2001:03:05 12:20)        //////
//////////////////////////////////////////////
// Methods of Slot:                    ///////
// - Slot                              ///////
// - ignoration                        ///////
// - getSlotFrame                      ///////
// - getName                           ///////
// - toString                          ///////
// - getFrameClassDeclaredIn           ///////
// - getType                           ///////
// - isAggType                         ///////
// - isUnorderedAggType                ///////
// - isOrderedAggType                  ///////
// - isOptional                        ///////
// - convertEmptyToNull                ///////
// - getIgnore                         //////
// - setIgnore                         /////
// - lookupInverseSlot                 ////
// - static writeClass                 ///
// - static readClass                  //
////////////////////////////////////////

import java.io.*;
import java.util.*;

import jex.*;

/*****************************************************************************
** The class `Slot' represents a slot in a frame. It has a name and a type.
** Each frame class has a set of slots associated with it, which instances of
** the frame class can have fillers for.
*****************************************************************************/

public class Slot implements Serializable {
	private Attribute _underlyingAttribute;
	private String _name;
	private FrameClass _frameClassDeclaredIn;
	private AbstractType _type;
	private static Hashtable/*String->
	               Hashtable<String->
	               Hashtable<FrameClass->Slot>>*/ _inverseSlots;
	private boolean _isAggType= false;
	private boolean _isOptional= false;
	private boolean _convertEmptyToNull= false;
	private boolean _ignore= false;

	/********************************************************************/
	// Slot constructor : Attribute ->
	/*********************************************************************
	** Constructs this slot by plugging in the given Jex Attribute
	** object which is taken to represent it (it does so by copying the
	** reference, which is maximally efficient, but the client should not
	** then alter the Attribute directly, as that would compromise the
	** integrity of the frame).
	** @param slotAttr A Jex Attribute object to wrap in this frame.
	** @param frameClass The frame class to which this slot belongs.
	** @param frameClassSchema The frame class schema in which this slot
	**  is defined.
	*********************************************************************/

	public Slot (Attribute slotAttr, FrameClass frameClass,
	             Schema frameClassSchema) {
		_underlyingAttribute= slotAttr;
		_frameClassDeclaredIn= frameClass;

		////////////////////
		// Explicit slots //
		////////////////////
		if (slotAttr instanceof ExplicitAttribute) {
			try {
				_name= ((ExplicitAttribute)slotAttr).getId();
				_type= ((ExplicitAttribute)slotAttr).
				                           getType();

				_isAggType= _type instanceof
				            AbstractAggregateTypes;
				_isOptional= ((ExplicitAttribute)slotAttr).
				             isOptional();
				_convertEmptyToNull= _isAggType && _isOptional
				 && ((IntegerLiteral)((AbstractAggregateTypes)
				    _type).getBoundSpec().getLowerBound().
				    getTerm().getFactor().getSimpleFactor1()).
				    getValue() > 0;
			}
			catch (UnsetValueException uve) {
				System.err.println("UnsetValueException "
				+ "caught in Slot constructor");
				System.exit(1);
			}
			catch (CloneNotSupportedException cnse) {
				System.err.println("CloneNotSupported"
				+ "Exception caught in Slot constructor");
				System.exit(1);
			}
			catch (JexException je) {
				System.err.println(je.getClass() + " caught "
				+ "in Slot constructor");
				System.err.println(je.getLocalizedMessage());
				je.printStackTrace();
				System.err.println("getType() failed for "
				+ "the explicit attribute "
				+ frameClass.getName() + ":" + _name);
				System.exit(1);
			}
		} else
		///////////////////
		// Inverse slots //
		///////////////////
		if (slotAttr instanceof InverseAttribute) {
			Entity entityInType= null;
			String thatFrameClass= null;
			String forAttrSlotName= null;
			try {
				_name= ((InverseAttribute)slotAttr).getId();

				_isAggType=
				   ((InverseAttribute)slotAttr).isSet()
				|| ((InverseAttribute)slotAttr).isBag();

				entityInType= (Entity)((InverseAttribute)
				  slotAttr).getEntityRef().
				            getReferencedObject();

				if (_inverseSlots == null)
					_inverseSlots= new Hashtable();

				thatFrameClass= entityInType.getId();

				forAttrSlotName=
				  ((InverseAttribute)slotAttr).
				  getAttributeRef().getId();
			}
			catch (UnsetValueException uve) {
				System.err.println("UnsetValueException "
				+ "caught in Slot constructor");
				System.err.println(uve.getMessage());
				uve.printStackTrace(System.err);
				System.exit(1);
			}
			catch (JexException je) {
				System.err.println(je.getClass() + " caught "
				+ "in Slot constructor");
				System.err.println(je.getLocalizedMessage());
				je.printStackTrace();
				System.err.println("getType() failed for "
				+ "the inverse attribute "
				+ frameClass.getName() + ":" + _name);
				System.exit(1);
			}

			String frameClassDeclaredIn= null;
			FrameClass thatFC= FrameClass.lookup
			                   (entityInType, frameClassSchema);
			try {
				frameClassDeclaredIn= thatFC.
					getSlot(forAttrSlotName).
					getFrameClassDeclaredIn().getName();
			} catch (InvalidSlot z) {
				System.err.println("Slot \"" + forAttrSlotName
				+ "\" not found in FrameClass \""
				+ thatFrameClass + "\"");
				System.exit(1);
			}
			Hashtable inverseSlots2= (Hashtable)
				_inverseSlots.get(frameClassDeclaredIn);
			if (inverseSlots2 == null) {
				inverseSlots2= new Hashtable();
				_inverseSlots.put(frameClassDeclaredIn,
				                  inverseSlots2);
			}
			Hashtable inverseSlots3= (Hashtable)
				inverseSlots2.get(forAttrSlotName);
			if (inverseSlots3 == null) {
				inverseSlots3= new Hashtable();
				inverseSlots2.put(forAttrSlotName,
				                  inverseSlots3);
			}
			inverseSlots3.put(frameClass,this);

			///////////////////////////////////////
			// The CCUS frame matching algorithm //
			// is to ignore inverse slots.       //
			///////////////////////////////////////
			_ignore= true;
		} else
		///////////////////
		// Derived slots //
		///////////////////
		if (slotAttr instanceof DerivedAttribute) {
			System.err.println
			           ("Derived slots not supported yet!");
		} else {
			System.err.println("This print statement could not " +
			                   "have been executed!!");
		}
	}

	/********************************************************************/
	// static Slot.ignoration : void -> void
	/*********************************************************************
	** Slots marked with --IGNORE in the frame class model are to be
	** ignored. This method is called from Frame.initialise.
	*********************************************************************/

	public static void ignoration () {
		BufferedReader fcm= null; FileInputStream fis= null;
		try {
			fcm= new BufferedReader(new InputStreamReader
			         (fis= new FileInputStream
			              ("EXPRESS/frame_classes.exp")));
		} catch (FileNotFoundException fnfe) {
			System.err.println("Slot.ignoration: FileNotFoundExce"
			+"ption for \"" + "EXPRESS/frame_classes.exp" + "\"");
			System.exit(1);
		}
		String line;
		String fcName= null; FrameClass frameClass= null;
		String slotName;
		int i;
		char c;
		try {
			while ((line= fcm.readLine()) != null) {
				try {
					if (line.substring(1,7).
					    equalsIgnoreCase("ENTITY"))
					{
						i= 8; // start of entity name
						fcName= "";
						while ((c=line.charAt(i))!=' '
						    && c != ';')
						{
							fcName+= c;
							i++;
						}
						frameClass
						= FrameClass.lookup(fcName);
					}
				} catch (IndexOutOfBoundsException ioobe) {}
				if (line.endsWith("--IGNORE")) {
					i= 2; // start of attribute name
					slotName= "";
					while ((c=line.charAt(i)) != ' ')
					{
						slotName+= c;
						i++;
					}
					try {
						frameClass.getSlot(slotName).
						           setIgnore(true);
					} catch (InvalidSlot is) {
						System.err.println("Slot.igno"
						+ "ration: InvalidSlot "
						+ fcName + "." + slotName);
					}
				}
			}
			fis.close();
		} catch (IOException ioe) {
			System.err.println("Slot.ignoration: IOException thro"
			+ "wn by BufferedReader.readLine()");
			System.exit(1);
		}
	}

	/********************************************************************/
	// Slot.getSlotFrame : void -> Frame
	/*********************************************************************
	** Delivers up the underlying frame of this slot.
	** @return The frame representing this slot.
	*********************************************************************/
 
	public Frame getSlotFrame () {
		return JexToFrames.getCachedFrame(_underlyingAttribute);
	}

	/********************************************************************/
	// Slot.getName : void -> String
	/*********************************************************************
	** Returns the name of this slot.
	** @return The name of this slot.
	*********************************************************************/
 	
	public String getName () {
		return _name;
	}
	
	/********************************************************************/
	// Slot.toString : void -> String
	/*********************************************************************
	** Returns a string representation of this slot.
	** @return The string representation of this slot.
	*********************************************************************/

	public String toString () {
		return _name;
	}

	/********************************************************************/
	// Slot.getFrameClassDeclaredIn : void -> String
	/*********************************************************************
	** Returns the frame class in which this slot was declared (the
	** supertype from which it is inherited).
	** @return The name of this slot.
	*********************************************************************/
 	
	public FrameClass getFrameClassDeclaredIn () {
		return _frameClassDeclaredIn;
	}
	
	/********************************************************************/
	// Slot.getType : void -> Object
	/*********************************************************************
	** Returns the type of this slot.
	** @return The type of this slot.
	*********************************************************************/
 
 	public Object getType () {
		try {
			return getSlotFrame().getSlotFiller("tipe");
		} catch (InvalidSlot x) {
			System.err.println("Something has gone very badly "
				+ "wrong for Slot.getType to throw an "
				+ "InvalidSlot exception!");
			return null;
		}
	}

	/********************************************************************/
	// Slot.isAggType : void -> boolean
	/*********************************************************************
	** Returns true just in case this slot is of an aggregation type (i.e.
	** to be filled by a set, bag, list or array).
	** @return Whether this slot is of an aggregation type.
	*********************************************************************/

	public boolean isAggType () {
		return _isAggType;
	}

	/********************************************************************/
	// Slot.isUnorderedAggType : void -> boolean
	/*********************************************************************
	** Returns true just in case this slot is of an unordered aggregation
	** type (i.e. to be filled by a set or bag).
	** @return Whether this slot is of an unordered aggregation type.
	*********************************************************************/

	public boolean isUnorderedAggType () {
		return _underlyingAttribute instanceof ExplicitAttribute
		    && (_type instanceof SetType || _type instanceof BagType)

		 ||    _underlyingAttribute instanceof InverseAttribute
		    && _isAggType;
	}

	/********************************************************************/
	// Slot.isOrderedAggType : void -> boolean
	/*********************************************************************
	** Returns true just in case this slot is of an ordered aggregation
	** type (i.e. to be filled by a list or array).
	** @return Whether this slot is of an ordered aggregation type.
	*********************************************************************/

	public boolean isOrderedAggType () {
		return _underlyingAttribute instanceof ExplicitAttribute
		 && (_type instanceof ListType || _type instanceof ArrayType);
	}

	/********************************************************************/
	// Slot.isOptional : void -> boolean
	/*********************************************************************
	** Returns true just in case this slot is optional (i.e. can have a
	** null filler).
	** @return Whether this slot is optional.
	*********************************************************************/

	public boolean isOptional () {
		return _isOptional;
	}

	/********************************************************************/
	// Slot.convertEmptyToNull : void -> boolean
	/*********************************************************************
	** Returns true just in case empty aggregations should be converted to
	** null to fill this slot.
	** @return Whether empty aggregations should be converted to null.
	*********************************************************************/

	public boolean convertEmptyToNull () {
		return _convertEmptyToNull;
	}

	/********************************************************************/
	// Slot.getIgnore : void -> boolean
	/*********************************************************************
	** Returns true just in case the CCUS frame matching algorithm should
	** ignore this slot.
	** @return True if the CCUS frame matching algorithm should ignore
	**         this slot, false otherwise.
	*********************************************************************/

	public boolean getIgnore () {
		return _ignore;
	}

	/********************************************************************/
	// Slot.setIgnore : boolean -> void
	/*********************************************************************
	** Marks this slot to be ignored (or not) by the CCUS frame matching
	** algorithm.
	** @param ignore True if the CCUS frame matching algorithm should
	**               ignore this slot, false otherwise.
	*********************************************************************/

	public void setIgnore (boolean ignore) {
		_ignore= ignore;
	}

	/********************************************************************/
	// Slot.lookupInverseSlot : (String x String x FrameClass) -> Slot
	/*********************************************************************
	** Looks up the inverse slot of a given slot in a given frame class
	** whose filler is frame of a given frame class.
	** @param frameClass The name of the frame class the slot whose
	** inverse is to be looked up is in.
	** @param slot The name of the slot whose inverse is to be looked up.
	** @param slotFillerFrameClass The frame class of the actual filler
	** (must be a frame) of `slot'.
	** @return The inverse slot of <i>frameClass</i>.<i>slot</i> for a
	** filler of type <i>slotFillerFrameClass</i>, or null if there isn't
	** one.
	*********************************************************************/

	public static Slot lookupInverseSlot (String frameClass, String slot,
	                               FrameClass slotFillerFrameClass)
	{
		if (_inverseSlots == null) return null;
		
		Hashtable inverseSlots2= (Hashtable)_inverseSlots.
		                                    get(frameClass);
		if (inverseSlots2 == null) return null;
		
		Hashtable inverseSlots3= (Hashtable)inverseSlots2.get(slot);
		if (inverseSlots3 == null) return null;

		Enumeration iS3Enum= inverseSlots3.keys();
		while (iS3Enum.hasMoreElements()) {
			FrameClass key= (FrameClass)iS3Enum.nextElement();
			if (key.subsumes(slotFillerFrameClass))
				return (Slot)inverseSlots3.get(key);
		}
		return null;
	}

	/********************************************************************/
	// static Slot.writeClass : ObjectOutput -> void
	/*********************************************************************
	** Serialises the static data members (class variables) of Slot.
	** @param oo An output stream to save the class data to.
	** @exception IOException If an I/O exception occurs.
	*********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_inverseSlots);
	}

	/********************************************************************/
	// static Slot.readClass : ObjectInputStream -> void
	/*********************************************************************
	** Deserialises the static data members (class variables) of Slot.
	** @param oi An input stream to save the class data to.
	** @exception IOException If an I/O exception occurs.
	*********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_inverseSlots= (Hashtable)oi.readObject();
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the st"
			+ "upid compiler is just too dumb to realise this!");
		}
	}

	/********************************************************************/
	// static Slot.showInverseSlots : void -> void
	/*********************************************************************
	** Shows the inverse slots lookup table - helpful for debugging.
	*********************************************************************/

	private static void showInverseSlots () {
		Enumeration inverseSlots= _inverseSlots.keys();
		System.err.println("++++++++++++++++++++++++++++++++++++++++"+
		                   "++++++++++++++++++++++++++++++++++++++++")
		;
		System.err.println("Slot._inverseSlots");
		while (inverseSlots.hasMoreElements()) {
			String frameClass= (String)inverseSlots.nextElement();
			System.err.println("\t" + frameClass + " ->");
			Hashtable inverseSlots2= (Hashtable)_inverseSlots.
			                                    get(frameClass);
			Enumeration iS2= inverseSlots2.keys();
			while (iS2.hasMoreElements()) {
				String slot= (String)iS2.nextElement();
				System.err.println("\t\t" + slot + " ->");
				Hashtable inverseSlots3= (Hashtable)
				          inverseSlots2.get(slot);
				Enumeration iS3= inverseSlots3.keys();
				while (iS3.hasMoreElements()) {
					FrameClass fc= (FrameClass)iS3.
					               nextElement();
					System.err.println("\t\t\t" + fc +
					" -> "+((Slot)inverseSlots3.get(fc)).
					       getName());
				}
			}
		}
		System.err.println("++++++++++++++++++++++++++++++++++++++++"+
		                   "++++++++++++++++++++++++++++++++++++++++")
		;
	}
}
